"""Trace implementation that stores traces in two files: JSON and HTML.

__WARNING__: This is an incubating feature. The trace format is subject to
changes and we do not guarantee backward compatibility at this stage.
"""

from __future__ import annotations

import asyncio
import base64
import datetime
import json
import os
from typing import Any, Tuple

from absl import logging
from genai_processors import content_api
from genai_processors.dev import trace
import pydantic
import shortuuid
from typing_extensions import override
import xxhash

HTML_TEMPLATE_PATH = os.path.join(os.path.dirname(__file__), 'trace.tpl.html')
with open(HTML_TEMPLATE_PATH, 'r') as f:
  HTML_TEMPLATE = f.read()

pydantic_converter = pydantic.TypeAdapter(Any)

_QUEUE_MAX_SIZE = 1000

_IMAGE_SIZE_KEY = 'image_size'


def _bytes_encoder(o: Any) -> Any:
  """Encodes bytes in parts based on mime type.

  The dump_python(mode='json') in Pydantic does not encode bytes in utf-8
  mode and this causes issues when sending ProcessorPart to JS/HTML clients
  (wrong padding, etc.). This function is used to handle bytes to base64
  encoding to match the behaviour of the JS/HTML side.

  Args:
    o: The object to encode.

  Returns:
    The encoded object.
  """
  if isinstance(o, bytes):
    return base64.b64encode(o).decode('utf-8')
  else:
    return pydantic_converter.dump_python(o, mode='json')


# TODO(elisseeff): Adjust the logic to make it less brittle. If a new bytes
# field is added to the ProcessorPart in the future this function will not
# decode it while it should.
def _bytes_decoder(dct: dict[str, Any]) -> Any:
  """Decodes base64 encoded bytes in parts based on mime type."""
  if 'data' in dct and 'mime_type' in dct and isinstance(dct['data'], str):
    mime_type = dct['mime_type']
    if not mime_type.startswith('text/') and not mime_type.startswith(
        'application/json'
    ):
      try:
        dct['data'] = base64.b64decode(dct['data'])
      except (ValueError, TypeError):
        pass
  return dct


def _resize_image_part(
    part: content_api.ProcessorPart, image_size: tuple[int, int]
) -> content_api.ProcessorPart:
  """Resizes image part."""
  try:
    img = part.pil_image
    img.thumbnail(image_size)
    metadata = part.metadata.copy() if part.metadata else {}
    metadata[_IMAGE_SIZE_KEY] = (img.width, img.height)
    part = content_api.ProcessorPart(
        img,
        role=part.role,
        substream_name=part.substream_name,
        mimetype=part.mimetype,
        metadata=metadata,
    )
  except Exception:  # pylint: disable=broad-except
    # If resizing fails, we just use the original part.
    pass
  return part


def _compute_part_hash(part_dict: dict[str, Any]) -> str:
  """Computes a hash for a part_dict for efficient deduplication.

  Args:
    part_dict: The part dictionary to hash.

  Returns:
    A hex string hash of the part_dict content.
  """
  json_str = json.dumps(part_dict, sort_keys=True, default=str)
  return xxhash.xxh128(json_str.encode('utf-8')).hexdigest()


class TraceEvent(pydantic.BaseModel):
  """A single event in a trace.

  An event represents an input/output part, an error, or a sub-trace from a
  nested processor call.
  """

  model_config = {'arbitrary_types_allowed': True}

  # A unique ID for this event.
  id: str = pydantic.Field(default_factory=lambda: str(shortuuid.uuid()))
  # The timestamp when the event was stored in the trace.
  timestamp: datetime.datetime = pydantic.Field(
      default_factory=datetime.datetime.now
  )
  # Whether the event is an input part to the processor or an output part.
  is_input: bool = False

  # Hash of the part (used for deduplication with the parts store in the root
  # trace).
  part_hash: str | None = None
  # If set, this event represents a nested processor call via its trace.
  sub_trace: SyncFileTrace | None = None
  # The relation between this trace and sub_trace. E.g. if it is a chain.
  relation: str | None = None
  # Error message if this event represents an error.
  error_message: str | None = None


class SyncFileTrace(trace.Trace):
  """A trace storing events in a file.

  The trace collects all events first in memory and then writes them to a file
  when the finalize method is called. Parts are deduplicated online as they are
  added, reducing memory footprint for traces with repeated content.
  """

  # Where to store the trace. Required only of the root trace.
  trace_dir: str | None = None

  # Maximum size of the trace in bytes.
  # If set on the root trace, parts added after the trace size exceeds this
  # limit will be stored with only their mimetype to save memory.
  max_size_bytes: int | None = None

  # The events in the trace. Collected in memory.
  events: list[TraceEvent] = pydantic.Field(default_factory=list)
  _queue: asyncio.Queue[Tuple[content_api.ProcessorPart, bool] | None] = (
      pydantic.PrivateAttr()
  )
  _worker: asyncio.Task[None] | None = pydantic.PrivateAttr(default=None)
  _size_limit_exceeded: bool = pydantic.PrivateAttr(default=False)
  _current_size_bytes: int = pydantic.PrivateAttr(default=0)

  # The size to resize images to when storing them in the trace.
  # If None, images are not resized.
  image_size: tuple[int, int] | None = (200, 200)

  # Parts store for deduplication: dictionary of a part's hash to the dict
  # representing the part. For sub-traces, this should be None; they use the
  # root's store.
  parts_store: dict[str, dict[str, Any]] | None = pydantic.Field(
      default_factory=dict
  )

  # Reference to root trace for sharing parts_store. None means this is root.
  _root_trace: 'SyncFileTrace | None' = pydantic.PrivateAttr(default=None)

  def _get_root(self) -> 'SyncFileTrace':
    """Returns the root trace for parts_store access."""
    return self._root_trace if self._root_trace is not None else self

  def _get_trace_filename(self) -> str:
    """Returns the filename for the trace."""
    return os.path.join(self.trace_dir, f'{self.name}_{self.trace_id}')

  def model_post_init(self, __context: Any) -> None:  # pylint: disable=invalid-name
    if self.trace_dir:
      logging.info('Start tracing to: %s', self._get_trace_filename())
    self._queue = asyncio.Queue(maxsize=_QUEUE_MAX_SIZE)
    self._worker = asyncio.create_task(self._event_worker())

  async def _event_worker(self):
    """Worker task to process parts from queue and create events."""
    while item := await self._queue.get():
      part, is_input = item
      await self._add_part(part, is_input=is_input)

  def to_json_str(self) -> str:
    """Converts the trace to a JSON string with deduplicated parts.

    The output format is:
    {
      "parts_store": {"0": {...}, "1": {...}, ...},
      "trace": { ... trace data with part_ref ... }
    }

    Returns:
      A JSON string representing the trace with deduplicated parts.
    """
    try:
      # We explicitly save parts store on the side to easily access it in the
      # json file.
      parts_store = self.parts_store
      trace_data = self.model_dump(
          mode='python',
          exclude_none=True,
          exclude={'parts_store'},
      )
      output = {
          'parts_store': parts_store,
          'trace': trace_data,
      }
      return json.dumps(
          output,
          default=_bytes_encoder,
          indent=2,
      )
    except TypeError as e:
      raise TypeError(
          'Failed to serialize trace to JSON. This might be due to'
          ' non-serializable types in ProcessorPart metadata. Ensure parts'
          ' added to traces are JSON-serializable.'
      ) from e

  @classmethod
  def from_json_str(cls, json_str: str) -> 'SyncFileTrace':
    """Initializes the trace from a JSON string.

    Args:
      json_str: The JSON string to initialize the trace from. The bytes field
        for audio and image parts are expected to be base64 + utf-8 encoded.

    Returns:
      The trace initialized from the JSON string. Note that the parts store
      cannot be reused for deduplication: a part added to the trace after
      deserialization will not be deduplicated against parts added before
      deserialization.
    """
    data = json.loads(json_str, object_hook=_bytes_decoder)
    parts_store = data['parts_store']
    trace_data = data['trace']
    trace_data['parts_store'] = parts_store
    trace_obj = cls.model_validate(trace_data)
    # Set _root_trace for all sub-traces
    cls._init_sub_traces(trace_obj, trace_obj)
    return trace_obj

  @classmethod
  def _init_sub_traces(
      cls, parent: 'SyncFileTrace', root: 'SyncFileTrace'
  ) -> None:
    """Recursively sets _root_trace for all sub-traces."""
    for event in parent.events:
      if event.sub_trace is not None:
        event.sub_trace._root_trace = root
        event.sub_trace.parts_store = None
        cls._init_sub_traces(event.sub_trace, root)

  async def _add_part(
      self, part: content_api.ProcessorPart, is_input: bool
  ) -> None:
    """Adds an input or output part to the trace events."""
    root = self._get_root()

    if (
        content_api.is_image(part.mimetype)
        and self.image_size
        and part.get_metadata(_IMAGE_SIZE_KEY, None) != self.image_size
        and not root._size_limit_exceeded
    ):
      # Resize image part if it is not already the desired size.
      part = await asyncio.to_thread(_resize_image_part, part, self.image_size)

    part_dict = part.to_dict(mode='python')
    # Remove capture_time field that is added by default by the GenAI Processor
    # library.
    part_dict['metadata'].pop('capture_time', None)
    part_hash = _compute_part_hash(part_dict)

    if root.parts_store is None:
      return None

    if part_hash in root.parts_store:
      pass
    elif root.max_size_bytes is None:
      root.parts_store[part_hash] = part_dict
    elif root._size_limit_exceeded:
      del part_dict['part']
      root.parts_store[part_hash] = part_dict
    else:
      part_json = json.dumps(part_dict, default=_bytes_encoder)
      part_size = len(part_json.encode('utf-8'))
      if root._current_size_bytes + part_size > root.max_size_bytes:
        logging.warning(
            'Trace size limit (%d bytes) exceeded. Skipping further content'
            ' for the rest of the trace (keep metadata and extra args only).',
            root.max_size_bytes,
        )
        root._size_limit_exceeded = True
        del part_dict['part']
      else:
        root._current_size_bytes += part_size
      root.parts_store[part_hash] = part_dict

    self.events.append(
        TraceEvent(
            part_hash=part_hash,
            is_input=is_input,
        )
    )

  @override
  async def add_input(self, part: content_api.ProcessorPart) -> None:
    """Adds an input part to the trace events."""
    await self._queue.put((part, True))

  @override
  async def add_output(self, part: content_api.ProcessorPart) -> None:
    """Adds an output part to the trace events."""

    await self._queue.put((part, False))

  @override
  def add_sub_trace(self, name: str, relation: str) -> SyncFileTrace:
    """Adds a sub-trace from a nested processor call to the trace events."""
    # This method must not block.
    t = SyncFileTrace(name=name, image_size=self.image_size)
    # Sub-traces share the root's parts_store for global deduplication
    t._root_trace = self._get_root()
    # Only the root trace has a parts_store.
    t.parts_store = None

    event = TraceEvent(sub_trace=t, is_input=False, relation=relation)
    self.events.append(event)
    return t

  @override
  async def add_error(self, error_message: str) -> None:
    """Adds an error event to the trace."""
    event = TraceEvent(error_message=error_message, is_input=False)
    self.events.append(event)

  @override
  def cancel(self) -> None:
    """Cancel the trace."""
    self.cancelled = True
    if self._worker:
      self._worker.cancel()
    if all(event.is_input or event.sub_trace for event in self.events):
      # If no output has been produced yet, clear the trace.
      self.events = []
      self.parts_store = {}

  @override
  async def _finalize(self) -> None:
    """Saves the trace to a file."""
    await self._queue.put(None)  # Sentinel to stop worker.
    await asyncio.shield(self._worker)

    indices_to_delete = []
    for i, event in enumerate(self.events):
      if sub_trace := event.sub_trace:
        await sub_trace._finalize()
        if not sub_trace.events:
          indices_to_delete.append(i)
    for i in reversed(indices_to_delete):
      # Update in place
      del self.events[i]

    if not self.trace_dir:
      return
    trace_filename = self._get_trace_filename()
    logging.info('Saving trace to: %s', trace_filename)
    await asyncio.to_thread(self.save, trace_filename + '.json')
    await asyncio.to_thread(self.save_html, trace_filename + '.html')

  def save_html(self, path: str) -> None:
    """Saves an HTML rendering of the trace to a file."""
    html = HTML_TEMPLATE.format(trace_json=self.to_json_str())
    with open(path, 'w') as html_file:
      html_file.write(html)
    logging.info('Saved HTML trace to: %s', path)

  def save(self, path: str) -> None:
    """Saves a trace to a file in JSON format.

    Args:
      path: The path to the file.
    """
    with open(path, 'w') as html_file:
      html_file.write(self.to_json_str())
    logging.info('Saved JSON trace to: %s', path)

  @classmethod
  def load(cls, path: str) -> 'SyncFileTrace':
    """Reads a trace from a JSON file.

    Args:
      path: The path to the file.

    Returns:
      The trace.
    """
    with open(path, 'r') as trace_file:
      return cls.from_json_str(trace_file.read())


SyncFileTrace.model_rebuild(force=True)



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Research Agent interfaces."""

import dataclasses
import dataclasses_json
from genai_processors import content_api
from google.genai import types as genai_types

ProcessorPart = content_api.ProcessorPart


@dataclasses_json.dataclass_json
@dataclasses.dataclass(frozen=True)
class Topic:
  """A topic to be researched.

  Attributes:
    topic: The topic to be researched.
    relationship_to_user_content: A description of how the topic is related to
      the user content.
    research_text: The research text for the topic. This field is optional and
      is expected to be populated by the TopicResearcher processor.
  """

  topic: str
  relationship_to_user_content: str
  research_text: str | None = None


@dataclasses.dataclass
class Config:
  """Config used by the Research agent.

  Attributes:
    topic_generator_model_name: The model name to use for the topic generator.
    topic_researcher_model_name: The model name to use for the topic researcher.
    research_synthesizer_model_name: The model name to use for synthesizing the
      research results.
    num_topics: The number of topics to generate.
    excluded_topics: A list of topics to exclude from the research. E.g., if the
      user prompt is "Starting a vegetable garden in my small backyard in
      London", then the excluded topics could be "what vegetables can grow in
      the London climate?", "vegetables that can grow with limited space", etc.
    enabled_research_tools: A list of research tools to enable. By default, only
      the Google Search tool is enabled.
  """

  topic_generator_model_name: str = 'gemini-2.5-flash'
  topic_researcher_model_name: str = 'gemini-2.5-flash'
  research_synthesizer_model_name: str = 'gemini-2.5-flash'
  num_topics: int = 5
  excluded_topics: list[str] | None = None
  # Use default factory instead
  enabled_research_tools: list[genai_types.ToolConfigOrDict] = (
      dataclasses.field(
          default_factory=lambda: [
              genai_types.Tool(google_search=genai_types.GoogleSearch())
          ]
      )
  )



import os
import shutil


def on_post_build(config, **kwargs):
  """Copy llms.txt to the site build directory."""
  base_dir = os.path.dirname(os.path.dirname(config['config_file_path']))
  src = os.path.join(base_dir, 'llms.txt')
  dest = os.path.join(config['site_dir'], 'llms.txt')
  if os.path.exists(src):
    shutil.copyfile(src, dest)



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Processor extracting parts from PDF documents (using pypdfium2).

Requires pypdfium2 to be installed:

```
pip install pypdfium2
```

See `pdf_cli.py` for an example of usage and to test this processor with a local
PDF file.
"""

import asyncio
from collections.abc import AsyncIterable
import io
import threading
from genai_processors import content_api
from genai_processors import processor
from PIL import Image
import pypdfium2 as pdfium


PDF_MIMETYPE = 'application/pdf'


def is_pdf(part: content_api.ProcessorPart) -> bool:
  return part.mimetype == PDF_MIMETYPE


def _png_from_pil(img: Image.Image) -> bytes:
  """Converts PIL Image to PNG bytes."""
  img_byte_arr = io.BytesIO()
  img.convert('RGB').save(img_byte_arr, format='png')
  return img_byte_arr.getvalue()


# PDFium is not thread-safe.
_PDFIUM_RENDERER_LOCK = threading.Lock()


class PDFExtract(processor.PartProcessor):
  """Extracts PDFs into content parts.

  Takes the PDF file bytes and extracts the text and images from it.
  The images are rendered as PNGs.
  The text is rendered as UTF-8.
  The PDF is rendered as a list of ProcessorParts.
  The PDF is rendered with pypdfium2.

  Note that the input parts with the PDF file bytes should also have a metadata
  field with the `original_file_name` key. This is used to generate the status
  messages.
  """

  def _format(
      self, file_name: str, pdf_doc: pdfium.PdfDocument
  ) -> tuple[content_api.ProcessorContent, tuple[str, int, int]]:
    total_pages_with_images = 0

    content = content_api.ProcessorContent()
    content += f'--- START OF PDF {file_name} ---\n\n'

    for i, page in enumerate(pdf_doc):
      # We first find out if there are any images on this page.
      # FPDF_PAGEOBJ_IMAGE is for images,
      # FPDF_PAGEOBJ_FORM is for images embedded as PDFs.
      images = page.get_objects(
          filter=(
              pdfium.raw.FPDF_PAGEOBJ_IMAGE,
              pdfium.raw.FPDF_PAGEOBJ_FORM,
              pdfium.raw.FPDF_PAGEOBJ_PATH,
          ),
          max_depth=0,
      )
      images = list(images)
      # If there are images on this page, we pass in a screenshot / render of
      # the page along with the raw text.
      if images:
        total_pages_with_images += 1
        screenshot_pil = page.render().to_pil()
        content += f'---- Screenshot for PAGE {i + 1} ----\n\n'
        content += content_api.ProcessorPart(screenshot_pil)
      content += f'--- PAGE {i + 1} ----\n\n'
      all_text_on_page = page.get_textpage().get_text_bounded()
      content += all_text_on_page
      page.close()

    return content, (file_name, len(pdf_doc), total_pages_with_images)

  def _process(
      self,
      part: content_api.ProcessorPart,
  ) -> tuple[content_api.ProcessorContent, tuple[str, int, int]]:
    # pdfium is not thread safe.
    with _PDFIUM_RENDERER_LOCK:
      file_name = part.get_metadata('original_file_name') or 'unknown.pdf'
      pdf_doc = pdfium.PdfDocument(part.bytes)
      try:
        return self._format(file_name, pdf_doc)
      finally:
        pdf_doc.close()

  def match(self, part: content_api.ProcessorPart) -> bool:
    return is_pdf(part)

  async def call(
      self, part: content_api.ProcessorPart
  ) -> AsyncIterable[content_api.ProcessorPart]:

    content, stats = await asyncio.to_thread(self._process, part)

    filename, pages, pages_with_images = stats
    yield processor.status(
        'Parsed PDF'
        f' {filename} ({pages} pages,'
        f' {pages_with_images} pages with images)'
    )

    for part in content.all_parts:
      yield part



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

r"""Live Commentator WebSocket server for AI Studio.

A server to run the live commentator agent on AI Studio. The applet connects
to this server and handles the UI.

See commentator.py for the actual implementation.

To run the server locally:

 * Install the dependencies with `pip install genai-processors`.
 * Access the applet at
 https://aistudio.google.com/app/apps/github/google-gemini/genai-processors/tree/main/examples/live/ais_app.
* Define a GOOGLE_API_KEY environment variable with your API key.
 * Launch the commentator agent: `python3 ./commentator_ais.py`.
 * Allow the applet to use a camera and enable one of the video sources.
"""

import asyncio
import os
from typing import Any

from absl import app
from absl import flags
from absl import logging
from genai_processors import processor
from genai_processors.examples import live_server
import commentator

_PORT = flags.DEFINE_integer(
    'port',
    8765,
    'Port to run this WebSocket server on.',
)
_DEBUG = flags.DEFINE_bool(
    'debug',
    False,
    'Enable debug logging.',
)
_TRACE_DIR = flags.DEFINE_string(
    'trace_dir',
    None,
    'If set, enable tracing and write traces to this directory.',
)


def create_live_commentator(
    config: dict[str, Any],
) -> processor.Processor:
  """Creates a live commentator processor."""
  chattiness = config.get('chattiness', 0.5)
  api_key = os.environ['GOOGLE_API_KEY']
  return commentator.create_live_commentator(
      api_key=api_key,
      chattiness=chattiness,
      unsafe_string_list=None,
  )


def main(argv):
  if len(argv) > 2:
    raise app.UsageError('Too many command-line arguments.')
  if _DEBUG.value:
    logging.set_verbosity(logging.DEBUG)
  asyncio.run(live_server.run_server(
      create_live_commentator,
      port=_PORT.value,
      trace_dir=_TRACE_DIR.value,
  ))


if __name__ == '__main__':
  app.run(main)



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Processors for local filesystem operations."""

import asyncio
import glob
import mimetypes
import os
import re
from typing import AsyncIterable

import aiofiles
from genai_processors import processor
from google.genai import types as genai_types


def _natural_sort_key(s: str) -> list[str | int]:
  """Key for natural sorting based on filename."""
  _, filename = os.path.split(s)
  return [
      int(text) if text.isdigit() else text.lower()
      for text in re.split("([0-9]+)", filename)
  ]


@processor.source(stop_on_first=False)
async def GlobSource(  # pylint: disable=invalid-name
    pattern: str,
    base_dir: str = ".",
    inline_file_data: bool = True,
) -> AsyncIterable[processor.ProcessorPartTypes]:
  """A source that yields files matching a glob pattern.

  Can be used to load context from files or to process a stream of them. In
  particular can be applied to process photos from a camera as they need to be
  treated as a stream to detect duplicates and select the best shots.

  Args:
    pattern: The glob pattern to match. This can be a path with a pattern.
    base_dir: The base directory to search from.
    inline_file_data: If True, the file content is loaded into the part.
      Otherwise, the part will contain a FileData object pointing to the file.
  """
  glob_pattern = os.path.join(base_dir, pattern)
  filenames = await asyncio.to_thread(glob.glob, glob_pattern, recursive=True)
  filenames.sort(key=_natural_sort_key)
  for filename in filenames:
    if os.path.isfile(filename):
      original_file_name = os.path.relpath(filename, base_dir)
      mimetype, _ = mimetypes.guess_type(filename)
      if not mimetype:
        raise ValueError(f"Could not guess MIME type of {filename!r}.")
      if inline_file_data:
        async with aiofiles.open(filename, "rb") as f:
          yield processor.ProcessorPart(
              await f.read(),
              mimetype=mimetype,
              metadata={"original_file_name": original_file_name},
          )
      else:
        yield processor.ProcessorPart(
            genai_types.Part(
                file_data=genai_types.FileData(
                    display_name=original_file_name,
                    file_uri=filename,
                    mime_type=mimetype,
                )
            ),
            mimetype=mimetype,
            metadata={"original_file_name": original_file_name},
        )



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Research agent example, using Processors library.

This agent takes user content as input, generates a set of topics based on
that content, researches each topic, and then synthesizes the research into a
final output.

The agent is composed of three main processors:
  - TopicGenerator: Generates `Topics` based on the user content.
  - TopicResearcher: Researches each `Topic`.
  - TopicVerbalizer: Converts a `Topic` ProcessorPart to a human-readable
  string.
  - The output of this flow is then passed to a GenAI model, which synthesizes
  the research into a final output.

The agent orchestrates these processors, passing the output of one to the
input of the next.
"""

import re

from . import agent
from . import interfaces
from .processors import topic_generator
from .processors import topic_researcher

ResearchAgent = agent.ResearchAgent
Config = interfaces.Config

TopicGenerator = topic_generator.TopicGenerator
TopicResearcher = topic_researcher.TopicResearcher

Topic = interfaces.Topic



# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A collection of agents that invoke underlying model in a way to improve quality."""

from typing import AsyncIterable

from genai_processors import content_api
from genai_processors import processor


class CriticReviser(processor.Processor):
  """Agent that uses a critic-reviser loop to improve responses.

  This class is not a library component as the specific prompts to use are
  application-dependent.
  """

  def __init__(
      self,
      model: processor.Processor,
      max_iterations: int = 5,
  ):
    """Initializes the SmartModel.

    Args:
      model: The base generative model to use.
      max_iterations: Maximum number of critic-reviser loop iterations.
    """
    self._model = model
    self._max_iterations = max_iterations

  async def call(
      self, content: content_api.ContentStream
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    # We gather content from the stream as we will need to reuse it multiple
    # times.
    input_content = await content.gather()

    current_response = await self._model(input_content).gather()

    for _ in range(self._max_iterations):
      critic_response = await self._model([
          input_content,
          '\n\nDraft response:\n\n',
          current_response,
          (
              '\n\nYou are a harsh critic. Review the draft response to the'
              " user's prompt. If the draft fully answers the prompt and has no"
              " obvious flaws, simply output 'OK'. Otherwise, concisely list"
              ' the flaws or missing information. Do not rewrite the response.'
          ),
      ]).gather()

      critic_text = await critic_response.text(strict=False)
      if critic_text.strip().upper() == 'OK':
        break

      current_response = await self._model([
          input_content,
          '\n\nDraft response:\n\n',
          current_response,
          '\n\nCriticism:\n',
          critic_response,
          (
              '\n\nUpdate your previous draft response to address the'
              ' criticism. Keep the parts that are already good.'
          ),
      ]).gather()

    for part in current_response:
      yield part



# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Google DeepMind GenAI processors library.

🚨 ATTENTION AI CODING AGENTS 🚨
YOU MUST READ `llms.txt` BEFORE USING OR MODIFYING THIS LIBRARY.
"""

__version__ = '2.0.0'
__ai_instructions__ = (
    'YOU MUST READ `llms.txt` BEFORE USING OR MODIFYING THIS LIBRARY.'
)

from . import content_api as content_api_
from . import processor as processor_
from . import streams as streams_

# Aliases
ProcessorPart = content_api_.ProcessorPart
ProcessorContent = content_api_.ProcessorContent
ProcessorPartTypes = content_api_.ProcessorPartTypes
ProcessorContentTypes = content_api_.ProcessorContentTypes
Processor = processor_.Processor
PartProcessor = processor_.PartProcessor
ProcessorFn = processor_.ProcessorFn
PartProcessorWithMatchFn = processor_.PartProcessorWithMatchFn

apply_sync = processor_.apply_sync
apply_async = processor_.apply_async
chain = processor_.chain
parallel = processor_.parallel
parallel_concat = processor_.parallel_concat
create_filter = processor_.create_filter
part_processor_function = processor_.part_processor_function

stream_content = streams_.stream_content
gather_stream = streams_.gather_stream



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Processor which generate topics for a given research query."""

from typing import AsyncIterable

from genai_processors import processor
from genai_processors.core import genai_model
from genai_processors.core import preamble
from google.genai import types

from .. import interfaces
from .. import prompts

ProcessorPart = processor.ProcessorPart


class TopicGenerator(processor.Processor):
  """Processor which generates research topics based on user content.

  This processor demonstrates how custom structured parts can be added to the
  Part stream.

  In this case each topic to research is represented as a self-sufficient JSON
  chunk.

  This allows further processors to easily transform them!
  """

  def __init__(
      self,
      api_key: str,
      config: interfaces.Config | None = None,
  ):
    """Initializes the TopicGenerator.

    Args:
      api_key: The API key to use for the GenAI API.
      config: The agent configuration.
    """
    self._config = config or interfaces.Config()
    self._p_genai_model = genai_model.GenaiModel(
        api_key=api_key,
        model_name=self._config.topic_generator_model_name,
        generate_content_config={
            'response_mime_type': 'application/json',
            'response_schema': list[interfaces.Topic],
        },
        http_options=types.HttpOptions(
            retry_options=types.HttpRetryOptions(
                attempts=100,
            ),
        ),
    )
    self._num_topics = self._config.num_topics

    preamble_content = [
        ProcessorPart(prompts.TOPIC_GENERATION_PREAMBLE),
        ProcessorPart(
            f'Please provide exactly {self._config.num_topics} research'
            ' topics, along with each topic\'s "relationship" to the user'
            ' prompt.'
        ),
    ]
    if self._config.excluded_topics:
      preamble_content.append(
          ProcessorPart(
              'Here is a list of topics that should be excluded:'
              f' {self._config.excluded_topics}'
          )
      )
    preamble_content.append(
        ProcessorPart('You will now be provided with the user content.')
    )
    p_preamble = preamble.Preamble(content=preamble_content)
    p_suffix = preamble.Suffix(
        content=[
            ProcessorPart(
                f"""Return your response as Topics JSON in the format below.

You MUST return exactly {self._config.num_topics} topics.

Topic
  topic: str
  relationship_to_user_content: list[str]

Topics
  list[Topic]

Your JSON:
"""
            )
        ]
    )

    self._pipeline = p_preamble + p_suffix + self._p_genai_model

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[ProcessorPart]:
    topics = []
    async for content_part in self._pipeline(content):
      topics.append(content_part.get_dataclass(interfaces.Topic))

    yield processor.status(f'Generated {len(topics)} topics to research!')
    for i, t in enumerate(topics):
      yield processor.status(
          f'Topic {i + 1}: "{t.topic}"\n\n*({t.relationship_to_user_content})*'
      )
      yield ProcessorPart.from_dataclass(t)



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Command Line Interface to run a simple live agent connected to the Live API.

## Setup

To install the dependencies for this script, run:

```
pip install --upgrade pyaudio genai-processors google-genai
```

Before running this script, ensure the `GOOGLE_API_KEY` environment
variable is set to the api-key you obtained from Google AI Studio.

Important: **Use headphones**. This script uses the system default audio
input and output, which often won't include echo cancellation. So to prevent
the model from interrupting itself it is important to use headphones.

## Run

To run the script:

```shell
python3 ./live_simple_cli.py
```

The script takes a video-mode flag `--mode`, this can be "camera", "screen". The
default is "camera". To share your screen run:

```shell
python3 ./live_simple_cli.py --mode=screen
```
"""


import argparse
import asyncio
import os

from absl import logging
from genai_processors.core import audio_io
from genai_processors.core import live_model
from genai_processors.core import text
from genai_processors.core import video
from google.genai import types as genai_types
import pyaudio

# You need to define the API key in the environment variables.
API_KEY = os.environ['GOOGLE_API_KEY']

INSTRUCTION_PARTS = [
    'You are an agent that interacts with the user in a conversation. Make'
    ' the conversation lively and interesting for the user. Approach the'
    ' conversation as if you were on TV anchor. You can make jokes, explain'
    ' interesting facts related to what you see and hear, predict what'
    ' could happen, judge some actions or reactions, etc. Respond to the'
    ' user in a few sentences maximum: keep it short and engaging. Avoid'
    ' long monologues. Feel free to commentate on anything you see. It is'
    ' up to you to decide to commentate or not. You can also ask questions'
    ' to the user. If you do not get any answer, you can ask again after a '
    ' while as in a normal conversation when the counterpart does not'
    ' listen or is distracted by something else.'
]


async def run_live(video_mode: str) -> None:
  r"""Runs a simple live agent taking audio/video streams as input.

  The audio and video input and output are connected to the local machine's
  default input and output devices.

  Args:
    video_mode: The video mode to use for the video. Can be CAMERA or SCREEN.
  """
  pya = pyaudio.PyAudio()
  video_mode_enum = video.VideoMode(video_mode)
  # input processor = camera/screen streams + audio streams
  # Note that the Live API requires audio/pcm mimetype (not audio/l16).
  input_processor = video.VideoIn(
      video_mode=video_mode_enum
  ) + audio_io.PyAudioIn(pya, use_pcm_mimetype=True)

  # Calls the Live API. If you define your own live agent, this is the processor
  # that will likely be replaced. See live/commentator_cli.py for a more
  # advanced example.
  live_processor = live_model.LiveProcessor(
      api_key=API_KEY,
      model_name='gemini-2.5-flash-native-audio-preview-12-2025',
      realtime_config=genai_types.LiveConnectConfig(
          system_instruction=INSTRUCTION_PARTS,
          # Ground with Google Search
          tools=[genai_types.Tool(google_search=genai_types.GoogleSearch())],
          # Return the transcription of the model audio.
          output_audio_transcription={},
          # Enable affective dialog (only available for native audio out models)
          enable_affective_dialog=True,
          response_modalities=['AUDIO'],
          # Set the language for the Live API.
          speech_config={'language_code': 'en-US'},
      ),
      http_options=genai_types.HttpOptions(api_version='v1alpha'),
  )

  # Plays the audio parts. This processor also handles interruptions and makes
  # sure the audio output stops when the user is speaking.
  play_output = audio_io.PyAudioOut(pya)

  # Creates an agent as: mic+camera -> Live API -> play audio
  live_agent = input_processor + live_processor + play_output

  print('Use ctrl+D to quit.')
  async for part in live_agent(text.terminal_input()):
    # Print the transcription and the output of the model (should include status
    # parts and other metadata parts)
    print(part)


if __name__ == '__main__':
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '--mode',
      type=str,
      default='camera',
      help='pixels to stream from',
      choices=['camera', 'screen'],
  )
  parser.add_argument(
      '--debug',
      type=bool,
      default=False,
      help='Enable debug logging.',
  )
  args = parser.parse_args()
  if not API_KEY:
    raise ValueError(
        'API key is not set. Define a GOOGLE_API_KEY environment variable with'
        ' a key obtained from AI Studio.'
    )
  if args.debug:
    logging.set_verbosity(logging.DEBUG)
  asyncio.run(run_live(video_mode=args.mode))



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Interface declaration for `ProcessorContent` cache."""

import abc
from collections.abc import Callable

from genai_processors import content_api


# Using CacheMiss = object() as a sentinel value doesn't play nicely with typing
class CacheMiss:
  """Sentinel value to represent a cache miss."""


CacheMissT = type[CacheMiss]

ProcessorContentTypes = content_api.ProcessorContentTypes
ProcessorContent = content_api.ProcessorContent


class CacheBase(abc.ABC):
  """Abstract base class for a cache for ProcessorContent."""

  @property
  @abc.abstractmethod
  def hash_fn(
      self,
  ) -> Callable[[ProcessorContentTypes], str | None]:
    """Returns the function used to convert a query into a lookup key."""

  @abc.abstractmethod
  async def lookup(
      self,
      query: ProcessorContentTypes | None = None,
      *,
      key: str | None = None,
  ) -> ProcessorContent | CacheMissT:
    """Looks up a ProcessorContent-like value in the cache for a given query."""

  @abc.abstractmethod
  async def put(
      self,
      *,
      query: ProcessorContentTypes | None = None,
      key: str | None = None,
      value: ProcessorContentTypes,
  ) -> None:
    """Puts a ProcessorContent value into the cache for a given query."""

  @abc.abstractmethod
  async def remove(self, query: ProcessorContentTypes) -> None:
    """Removes a value from the cache for a given query."""

  @abc.abstractmethod
  def with_key_prefix(self, prefix: str) -> 'CacheBase':
    """Creates a new Cache instance where generated string keys are prefixed."""



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Appends a preamble to a stream."""

from collections.abc import AsyncIterable, Callable

from genai_processors import content_api
from genai_processors import processor

PreambleFactory = Callable[[], content_api.ProcessorPartTypes]


class Preamble(processor.Processor):
  """Prepends a preamble to the full content."""

  def __init__(
      self,
      *,
      content: content_api.ProcessorContentTypes | None = None,
      content_factory: PreambleFactory | None = None,
  ):
    """Constructs a Preamble processor.

    Args:
      content: content to prepend.
      content_factory: function for returning a content given no input. This is
        helpful for when contents are not fully known on __init__, e.g. if they
        depend on the user or time of the request.

    Raises:
      ValueError if both `content` and `content_factory` are provided.
    """
    if content is not None and content_factory is not None:
      raise ValueError(
          'Only one of `content` and `content_factory` must be provided.'
      )

    self._content = (
        None if content is None else content_api.ProcessorContent(content)
    )
    self._content_factory = content_factory

  async def call(
      self, content: AsyncIterable[content_api.ProcessorPart]
  ) -> AsyncIterable[content_api.ProcessorPart]:
    if self._content is not None:
      for content_part in self._content:
        yield content_part
    elif self._content_factory is not None:
      for content_part in content_api.ProcessorContent(self._content_factory()):
        yield content_part

    async for content_part in content:
      yield content_part


class Suffix(Preamble):
  """Appends a suffix to the full content."""

  async def call(
      self, content: AsyncIterable[content_api.ProcessorPart]
  ) -> AsyncIterable[content_api.ProcessorPart]:
    async for content_part in content:
      yield content_part

    if self._content is not None:
      for content_part in self._content:
        yield content_part
    elif self._content_factory is not None:
      for content_part in content_api.ProcessorContent(self._content_factory()):
        yield content_part



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Function calling processor.

The FunctionCalling processor is a loop over a model call that intercepts any
function call, runs the function, injects the result back into the model prompt
and goes to the next iteration until no function call is made.

The term function here is a bit of a misnomer as not all functions can be
called: its parameters must be JSON-serializable, see
https://ai.google.dev/gemini-api/docs/function-calling?example=meeting#schema.
Its responses should also either be JSON-serializable values or should return a
genai_types.FunctionResponse. We will therefore refer to such functions as tools
from now on to avoid confusion with functions defined in Python.

We define two types of tools: sync and async. Both can be easily defined from
sync of async Python functions.

For sync tools, the function calling processor will wait until the function
returns the result. This is the default behavior when the model is turn-based.
When the model is bidi (aka realtime), the function will be run in a separate
thread and any result will trigger an interruption of the model output and a new
model call.

If you provide an async tool, the function calling processor will return a 'OK,
running in the background.' response to the model immediately and will send back
the result later on when it is available. Such behavior happens for turn-based
models as well as for realtime models. For turn-based models, the model will
wait for the async tools to return the result before returning to the user.
If the model outputs more than one async function/tool call, these tools will
be run in parallel and the results can be interleaved. The function responses
contain the name of the tool that returned the response. This allows you to
detect which response corresponds to which tool call.

We therefore have the following 4 cases:

- Turn-based model + sync tool: tool is blocking future model calls. It is run
in a separate thread. Many can be run concurrently if the model output more than
one tool call at once. The model will wait for all tools to finish before
returning.
- Turn-based model + async tool: tool is non-blocking, allowing the model to
continue returning tool calls in subsequent model calls. The model will wait for
all tools to finish before returning control to the user.
- Realtime model + sync tool: tool is automatically converted to an async tool.
to let the realtime model process new incoming input without being blocked.
- Realtime model + async tool: tool is non-blocking, many can be run in
parallel and async tools can stream responses back to the model as they become
available.

For async tools and more generally for any tool used in with a realtime model,
you will get a first and immediate 'OK, running in the background.' response
containing a tool call ID, which is needed to associate future function
responses to the  correct calls. For sync tools, you will get the result as a
single function response when the tool returns. It is scheduled just after the
tool call.

We recommend using sync tools for turn-based models as they have
simpler formatting: they don't need the 'OK, running in the background.'
response with a tool call ID as the function response is returned after the tool
call.

When using a realtime model, sync tools are wrapped into a thread and de facto
becomes async to avoid blocking the realtime model. This means the model will
not wait for the tool to finish before returning and can process new input while
the tool is running. This means any tool used in a realtime model will have
a first 'OK, running in the background.' immediate response containing a tool
call ID.

Function calling processor provides a function to cancel any async tool. This
cancellation is implemented by the function calling processor directly. You
still need to provide the function declaration to the model in its prompt to
make it aware that this function is available. See below how to add the
cancel function to the model prompt. If you do not provide this cancel function
to the model, function calling will still work but the user will not be able to
stop an async tool.

You can define an async tool from a Python async function or from a Python async
generator. Async generators will stream responses back to the model whenever it
can. This can be used for proactive applications, e.g. geofence tool can yield a
response every time a user enters or exits their home, thus waking an assistant
and giving it a chance to react.

The response of async tools are automatically wrapped into
`genai_types.FunctionResponse`. This is what is returned by the function calling
processor. To control how the function response is added to the model prompt,
you can return `FunctionResponse` directly from your async tool function.

The `FunctionResponse` object includes a scheduling option that lets you control
how the function response is added to the model prompt:

- SILENT: the response is added to the model prompt but no model output is
generated. The model will see it next time something else triggers the
generation.
- INTERRUPT: the response is added to the model prompt, and the model output is
generated interrupting any ongoing model output.
- WHEN_IDLE: we wait for the model to output its current result (if there is
an ongoing one) then the response is appended to the prompt and the model is
requested to generate a response. This is the default option.

Note that you can return FunctionResponse in an async generator function.

Any async tool (or sync tool in a bidi model) called by the model will trigger
the following `FunctionResponse`. This is done automatically by this function
calling processor:

```python
FunctionResponse(
    name='<tool_name>',  # name of the tool or of the python function
    function_call_id='<id provided by the function call>',
    response={'result': 'Running in background.'},
    role='user',
    substream_name=<function_call_substream_name>, # provided to the ctor.
    scheduling='SILENT', # does not trigger any model output. just for info.
)
```

The function calling processor will populate any function ouptut with the proper
tool call ID and substream name, even when the function returns a
`FunctionResponse` directly.

This function calling processor focuses on the function calling loop and on the
execution of tools only. The model is responsible for formatting/parsing these
parts into its internal representation.

The overall logic can be depicted as:

```
    input -> pre_processor -> model -----+----> output
                  ^                      |
                  |                      v
            function run ----<---- function call
```

where:
-  `input` is the input stream.
-  `pre_processor` is a processor called once on `input` and on all results
obtained by tool invocation. This could for instance be an image tokenizer to
preprocess images and avoid re-tokenizing them on every model invocation.
-  `model` is a model processor that generates content. For the model to know
which tools are available, the samefunction/tool set should be given both to the
FunctionCalling and the model constructor. This model can be turn-based or
bidi but this should be specificed in the FunctionCalling ctor.
-  `function_call` executes the function calls returned by the model. This is
a private processor in this file. The function response is then fed back to the
model for another iteration. If there is no function call to execute and the
the model is turn-based, the loop is stopped and the current tool_use output is
sent back to the output stream.

The output is the model output including the function call and the responses.
They are all in the same substream as identified by `substream_name` in the
FunctionCalling ctor.

When used with GenAI models (i.e. Gemini API), the model should be defined as
follows:

```python
genai_processor = genai_model.GenaiModel(
    api_key=API_KEY,
    model_name='gemini-2.5-flash',
    generate_content_config=genai_types.GenerateContentConfig(
        tools=[fns],
        automatic_function_calling=genai_types.AutomaticFunctionCallingConfig(
            disable=True
        ),
    ),
)
```

where `fns` are the python functions to be called. Note that we disable the
automatic function calling feature here to avoid duplicate function calls with
the GenAI automatic function calling feature.

If you want to allow cancelling ongoing async functions add `cancel_fc` and
`list_fc` tool defined in this file:

```python
generate_content_config=genai_types.GenerateContentConfig(
    tools=[fns, function_calling.cancel_fc, function_calling.list_fc],
    automatic_function_calling=genai_types.AutomaticFunctionCallingConfig(
        disable=True
    ),
)
```

You define the function `fns` as regular python functions. Arguments and
outputs need to be JSON-serializable for models to be able to generate and
ingest them:

```python
def get_weather(location: str, period: Tuple[str, str] | None = None) -> str:
  '''Returns the weather for the provided period of time.

  Args:
    location: The location to get the weather for.
    period: A (start_time, end_time) tuple.

  Returns:
    A string describing the weather for the provided periods of time. If no
    period is provided, the current weather is returned.
  '''
  return 'Sun is shining'
```

We provide an extensive docstring here as this is usually used by the model to
add function declarations to its prompt. Supplying async Python functions
directly can be done as follows:

```python
async get_weather(location: str) -> str:
  '''Returns the current weather at the provided location.

  Args:
    location: The location to get the weather for.

  Returns:
    A string describing the weather for the provided location at the current
    time.
  '''
  response = await httpx.AsyncClient(...).get()
  return response.json()['temperature']
```

This creates an async tool automatically: the function calling processor will
return a 'OK, running in the background.' response to the model immediately and
will send back the result later on when it is available.

When an async function returns more than one response (over a longer period of
time), it is recommended to define an async generator function:

```python
async set_timer(countdown_seconds: int) -> AsyncIterable[str]:
  '''Sets a timer for the given number of seconds.

  Args:
    countdown_seconds: The number of seconds to set the timer for.

  Yields:
    A string describing the progress of the timer every second.
  '''
  await asyncio.sleep(countdown_seconds)
  yield f'{countdown_seconds} seconds have passed.'
```
  count_tick = 0
  while (count_tick < countdown_seconds)
    await asyncio.sleep(1)
    count_tick += 1
    yield f'{count_tick} seconds have passed.'

The function calling processor will detect async generators and will stream
responses back to the model whenever it can. This can be used for proactive
applications when an async generator function for instance sends signals to the
model every time it detects a change in a video feed.

When the end of the async generator is reached, the async tool will return an
empty `FunctionResponse` with the `will_continue` field set to `False` and the
scheduling set to `SILENT` (for info to the model). All other function responses
in the async generator are returned with the `will_continue` field set to
`True` to indicate that the tool is continuing to run and might produce more
responses.

As discussed above, you can return FunctionResponse directly from your async
tool function or async generator function to control how the function response
is added to the model prompt.
"""

import asyncio
import collections
from collections.abc import AsyncIterable
import dataclasses
import inspect
import math
from typing import Any, Callable

from genai_processors import content_api
from genai_processors import context as context_lib
from genai_processors import mcp
from genai_processors import processor
from genai_processors import streams
from google.genai import _extra_utils
from google.genai import types as genai_types

# All function call parts (calls and responses) are emitted in a substream
# with this name. This is to help downstream processors to identify function
# calls that were executed from function calls returned directly by the model.
FUNCTION_CALL_SUBSTREAM_NAME = 'function_call'

# Internal metadata key to carry the scheduling of a function call in an
# EndOfTurn part.
_SCHEDULING_METADATA_KEY = '__scheduling__'


def _get_scheduling(
    part: content_api.ProcessorPart,
) -> genai_types.FunctionResponseScheduling:
  if part.function_response and part.function_response.scheduling:
    return part.function_response.scheduling
  return part.get_metadata(
      _SCHEDULING_METADATA_KEY,
      genai_types.FunctionResponseScheduling.WHEN_IDLE,
  )


@dataclasses.dataclass
class _FunctionCallState:
  """State of the function calling processor."""

  # Number of async function calls that are currently running.
  running_fc_count: int = 0

  # Whether the model is currently outputting.
  model_outputting: bool = True
  # Number of tool calls that have been made.
  fn_call_count: int = 0
  # Maximum number of tool calls to make.
  fn_call_count_limit: float = math.inf
  # Whether a model call should be scheduled after an async function sent back
  # a part to the model prompt.
  schedule_model_call: bool = False


def _to_bidi(p: processor.Processor) -> processor.Processor:
  """Converts a unary processor to a bidi processor.

  A bidi (aka realtime) processor has actually more features than what we return
  here: eager processing of input, to handle start/end of speech signals, etc.
  Here, we need a thin adapter from turn-based to bidi using EoT to trigger a
  model call. This is to keep the same function calling loop between bidi and
  non-bidi so the logic stays the same as much as possible.

  Args:
    p: The unary processor to convert.

  Returns:
    A processor that can take an infinite input stream and will trigger model
    calls whenever an EndOfTurn is received.
  """

  @processor.processor_function
  async def call(
      content: AsyncIterable[content_api.ProcessorPart],
  ) -> AsyncIterable[content_api.ProcessorPart]:
    prompt = []
    dirty_prompt = False
    async for part in content:
      if content_api.is_end_of_turn(part):
        dirty_prompt = False
        async for part in p(streams.stream_content(prompt)):
          yield part
        yield content_api.END_OF_TURN
      else:
        prompt.append(part)
        if part.role == 'user':
          dirty_prompt = True
    if dirty_prompt:
      async for part in p(streams.stream_content(prompt)):
        yield part
      yield content_api.END_OF_TURN

  return call


async def _add_end_of_turn(
    content: AsyncIterable[content_api.ProcessorPart],
) -> AsyncIterable[content_api.ProcessorPart]:
  """Adds an end of turn to the content if it is not already present."""
  last_part = None
  async for part in content:
    yield part
    last_part = part
  if last_part is None or not content_api.is_end_of_turn(last_part):
    yield content_api.END_OF_TURN


class FunctionCalling(processor.Processor):
  """Tool use with Function Calling.

  See file level docstring for more details.
  """

  def __init__(
      self,
      model: processor.Processor,
      *,
      is_bidi_model: bool = False,
      substream_name: str = FUNCTION_CALL_SUBSTREAM_NAME,
      pre_processor: (
          processor.Processor | processor.PartProcessor | None
      ) = None,
      fns: (
          list[Callable[..., Any] | genai_types.McpClientSession] | None
      ) = None,
      max_function_calls: int | None = None,
  ):
    """Initializes the FunctionCalling processor.

    Args:
      model: The processor to use for generation. The model can be a bidi or
        unary model. If it is a unary model, the function calling processor will
        convert it to a bidi processor. If this is a realtime model like the
        Gemini Live API or like the `realtime.LiveProcessor`, `is_bidi_model`
        should be set to True.
      is_bidi_model: Whether the model is a bidi model. Most Gemini API models -
        except the Live API - are unary streaming, i.e. `is_bidi_model` is
        False. If `is_bidi_model` is False, the function calling processor will
        convert the unary processor to a bidi processor that can be called with
        async function calls. This means the model will wait for the end of any
        async function called before returning, the TTFT is therefore bounded by
        the longest async function called. With true bidi models, the model
        stops whenever the input stream ends.
      substream_name: The substream name to use for function calls and function
        responses. This is to help downstream processors to identify function
        calls and responses that were executed from this function calling
        processor.
      pre_processor: An optional pre-processor to pass the model input (prompt,
        function responses, model output from previous iterations) through.
      fns: The functions to register for function calling. Those functions must
        be known to `model`, and will be called only if `model` returns a
        function call with the matching name. For Gemini API, this means the
        same functions should be passed in the `GenerationConfig(tools=[...])`
        to the `model` constructor. If the function name is not found in the
        `fns` list, the function calling processor will return the unknown
        function call part and will raise a `ValueError`. If the execution of
        the function fails, the function calling processor will return a
        function response with the error message. MCP tools are automatically
        converted to callables when passed in as functions to this processor,
        this happens the first time the model is called.
      max_function_calls: maximum number of function calls to make (default set
        to 5 of turn-based models and infinity for bidi aka realtime models).
        When this limit is reached, the function calling loop will wait for the
        last function calls to return. If a function call is async and runs for
        a long time, the function calling loop will wait for it to finish.
    """
    self._model = model
    self._substream_name = substream_name
    self._is_bidi_model = is_bidi_model
    self._fns = fns
    self._fns_initialized = False
    self._pre_processor = (
        pre_processor.to_processor()
        if pre_processor
        else processor.passthrough().to_processor()
    )
    default_max_function_calls = 5 if not is_bidi_model else math.inf
    self._max_function_calls = max_function_calls or default_max_function_calls

  async def _initialize_mcp_tools(self):
    """Initializes the MCP tools."""
    if self._fns_initialized:
      return
    fns = []
    for fn in self._fns:
      if isinstance(fn, genai_types.McpClientSession):
        await fn.initialize()
        fns.extend(await mcp.mcp_tools_to_callables(fn))
      else:
        # Keep the existing functions.
        fns.append(fn)
    self._fns_initialized = True
    self._fns = fns

  async def call(
      self, content: AsyncIterable[content_api.ProcessorPart]
  ) -> AsyncIterable[content_api.ProcessorPart]:
    await self._initialize_mcp_tools()
    state = _FunctionCallState(fn_call_count_limit=self._max_function_calls)
    # To support both bidi and unary models we reduce both cases to bidi.
    model = _to_bidi(self._model) if not self._is_bidi_model else self._model
    # We use input_queue to feed function responses to consecutive turns.
    input_queue = asyncio.Queue[content_api.ProcessorPart | None]()
    input_stream = streams.merge(
        _add_end_of_turn(content),
        streams.dequeue(input_queue),
        stop_on_first=self._is_bidi_model,
    )

    output_queue = asyncio.Queue[content_api.ProcessorPart | None]()
    execute_function_call = _ExecuteFunctionCall(
        fns=self._fns,
        output_queue=output_queue,
        fn_state=state,
        is_bidi_model=self._is_bidi_model,
        substream_name=self._substream_name,
    )

    async def processor_pipeline():
      async def preprocess():
        async for part in self._pre_processor(input_stream):
          if content_api.is_end_of_turn(part):
            state.model_outputting = True
          yield part

      try:
        async for part in model(preprocess()):
          # Models don't always set the role and content_api.END_OF_TURN is
          # explicitly set to role='user'.
          if not part.role or content_api.is_end_of_turn(part):
            part.role = 'model'
          if part.function_call:
            await execute_function_call(part)
          else:
            await output_queue.put(part)
      finally:
        await output_queue.put(None)

    pipeline_task = processor.create_task(processor_pipeline())

    deferred_function_responses = []

    async def _yield_part(part):
      if not self._is_bidi_model or part.function_response:
        await input_queue.put(part)
      yield part

      if part.function_call:
        state.running_fc_count += 1

      if part.function_response:
        if not part.function_response.will_continue:
          state.running_fc_count -= 1

        match _get_scheduling(part):
          case genai_types.FunctionResponseScheduling.SILENT:
            pass
          case genai_types.FunctionResponseScheduling.WHEN_IDLE:
            if state.model_outputting:
              state.schedule_model_call = True
            else:
              await input_queue.put(content_api.END_OF_TURN)
          case _:
            await input_queue.put(content_api.END_OF_TURN)
            # If we reached max count, we should allow the model to react to
            # this response, so we make one extra iteration and stop after it.
        if state.fn_call_count >= state.fn_call_count_limit + 1:
          await input_queue.put(None)
          await output_queue.put(None)
          pipeline_task.cancel()

    async for part in streams.dequeue(output_queue):
      if content_api.is_end_of_turn(part) and part.role == 'model':
        state.model_outputting = False

        had_non_silent_deferred = False
        for deferred in deferred_function_responses:
          if deferred.function_response:
            if (
                _get_scheduling(deferred)
                != genai_types.FunctionResponseScheduling.SILENT
            ):
              had_non_silent_deferred = True
          async for yielded in _yield_part(deferred):
            yield yielded
        deferred_function_responses.clear()

        # EoT issued by the model. We need to check if we stop or if there is
        # a function call to schedule.
        if not self._is_bidi_model:
          if state.schedule_model_call:
            await input_queue.put(content_api.END_OF_TURN)
            state.schedule_model_call = False
          elif not had_non_silent_deferred and state.running_fc_count == 0:
            # Stop when there are no more function calls to execute and no more
            # async function calls running, or when the max function calls is
            # reached.
            await input_queue.put(None)
            await output_queue.put(None)
        else:
          if state.schedule_model_call and not state.model_outputting:
            input_queue.put_nowait(content_api.END_OF_TURN)
            state.schedule_model_call = False
          elif state.running_fc_count == 0:
            yield part
      else:
        if context_lib.is_reserved_substream(part.substream_name):
          yield part
        else:
          should_defer = False
          if part.function_response:
            should_defer = _get_scheduling(part) in (
                genai_types.FunctionResponseScheduling.SILENT,
                genai_types.FunctionResponseScheduling.WHEN_IDLE,
            )
          if should_defer and state.model_outputting:
            deferred_function_responses.append(part)
          else:
            async for yielded in _yield_part(part):
              yield yielded


async def cancel_fc(function_ids: list[str]) -> content_api.ProcessorPart:
  """Cancels function calls whose ids are in the provided argument.

  Args:
    function_ids: The list of ids of the function calls to cancel.

  Returns:
    A text message stating whether the functions have been successufully
    cancelled or if there were some errors e.g. function with such id does not
    exist.
  """
  del function_ids
  raise NotImplementedError(
      'This cancel_fc is an interface without implementation that can be used'
      ' as a tool definition for models. Actual implementation is in'
      ' FunctionCalling.cancel_fc.'
  )


def list_fc() -> str:
  """Returns the list of function calls that are currently running.

  A function call is represented as a string describing the function name, its
  arguments, and its id.

  Returns:
    A text paragraph with one item per function call, each item describing the
    function call that were made and are currently running.
  """
  raise NotImplementedError(
      'This list_fc is an interface without implementation that can be used'
      ' as a tool definition for models. Actual implementation is in'
      ' FunctionCalling.list_fc.'
  )


class _ExecuteFunctionCall:
  """Executes a function call and returns the whole input with the result."""

  def __init__(
      self,
      fns: list[Callable[..., Any]],
      output_queue: asyncio.Queue[content_api.ProcessorPart | None],
      fn_state: _FunctionCallState,
      is_bidi_model: bool,
      substream_name: str,
  ):
    self._fns = {fn.__name__: fn for fn in fns}
    if is_bidi_model:
      # Cancel and list functions make only sense for bidi models.
      self._fns['cancel_fc'] = self.cancel_fc
      self._fns['list_fc'] = self.list_fc
    self._output_queue = output_queue
    # Map from function call id to the task running the function call.
    self._task_map: dict[str, asyncio.Task] = {}
    self._call_map: dict[str, genai_types.FunctionCall] = {}
    self._task_counter = collections.defaultdict(int)
    self._fn_state = fn_state
    self._is_bidi_model = is_bidi_model
    self._substream_name = substream_name

  def _create_task_id(self, fn_name: str) -> str:
    """Creates a function call id."""
    # Semantic task IDs help models to keep track of running tools and reduce
    # halucinations.
    task_id = f'{fn_name}_{self._task_counter[fn_name]}'
    self._task_counter[fn_name] += 1
    return task_id

  def list_fc(self) -> content_api.ProcessorPart:
    """Returns a list of function calls that are currently running."""
    f_calls = ['Background Functions currently running:']
    for k, v in self._task_map.items():
      if not v.done() and 'list_fc' not in k:
        f_calls.append(
            f'Function {self._call_map[k].name} is running with args'
            f' {self._call_map[k].args} and id: {k}'
        )
    return content_api.ProcessorPart.from_function_response(
        name='list_fc',
        response='\n'.join(f_calls) + '\n',
        role='user',
        substream_name=FUNCTION_CALL_SUBSTREAM_NAME,
    )

  async def cancel_fc(
      self, function_ids: list[str]
  ) -> content_api.ProcessorPart:
    """Cancels function calls whose ids are in the provided argument."""
    cancelled_tasks = []
    for function_id in function_ids:
      if function_id in self._task_map:
        if self._task_map[function_id].cancel():
          try:
            await self._task_map[function_id]
          except asyncio.CancelledError:
            cancelled_tasks.append(function_id)
        else:
          # When the task is already done, we assume cancellation was
          # successful.
          cancelled_tasks.append(function_id)
    diff = set(function_ids) - set(cancelled_tasks)
    if not diff:
      response = 'OK, cancelled.'
      is_error = False
    elif cancelled_tasks:
      response = (
          f'Cancelled the following function calls: {cancelled_tasks}.'
          f' The following function calls were not found: {diff}.'
      )
      is_error = True
    else:
      response = (
          f'Could not find any of the following function calls: {function_ids}.'
      )
      is_error = True
    return content_api.ProcessorPart.from_function_response(
        name='cancel_fc',
        response=response,
        role='user',
        substream_name=self._substream_name,
        is_error=is_error,
        scheduling='SILENT',
    )

  def _add_task(
      self, call: genai_types.FunctionCall, fn: Callable[..., Any]
  ) -> None:
    """Adds the Task running a function and returns task_id."""
    task_id = call.id
    task = processor.create_task(
        self._put_function_response_to_output_queue(call, fn)
    )
    self._task_map = {k: v for k, v in self._task_map.items() if not v.done()}
    self._task_map[task_id] = task
    self._call_map[task_id] = call
    self._call_map = {k: self._call_map[k] for k in self._task_map.keys()}

  def _to_function_response(
      self,
      parts: content_api.ProcessorContentTypes | Any,
      call: genai_types.FunctionCall,
      will_continue: bool | None = None,
      scheduling: genai_types.FunctionResponseScheduling | None = None,
  ) -> content_api.ProcessorPart:
    """Wraps the part in a FunctionResponse if it is not already one.

    Tools in function calling do not always return a FunctionResponse. This
    method wraps the part in a FunctionResponse if it is not already one. If it
    is already a FunctionResponse, we update the id to match the function call
    id.

    The response is structured as {'result': str(part)}. Part must therefore be
    convertible to a string (e.g. json serializable).

    Args:
      parts: The parts to wrap. You can pass more than one part. It will be
        wrapped in a single function response. When providing something
        different than a `ProcessorContentTypes`, `parts` needs to be JSON
        serializable. When providing a function response, the id will be set to
        the function call id and the substream name will be set to the function
        calling processor substream name (changed in place).
      call: The function call that was made.
      will_continue: should only be used for async generator functions. If set,
        will override any provided value. Otherwise, the existing value is
        preserved.
      scheduling: The scheduling of the function response. If set, will override
        any provided value. Otherwise, the existing value is preserved.

    Returns:
      A function response wrapping the content in the part.
    """
    substream_name = self._substream_name
    if isinstance(parts, content_api.ProcessorPart) and parts.substream_name:
      substream_name = parts.substream_name

    if isinstance(parts, content_api.ProcessorPart) and parts.function_response:
      parts.function_response.id = call.id
      if not parts.function_response.name:
        parts.function_response.name = call.name
      if not parts.substream_name:
        parts.substream_name = self._substream_name
      if not parts.role:
        parts.role = 'user'
      if will_continue is not None:
        parts.function_response.will_continue = will_continue
      if scheduling is not None:
        parts.function_response.scheduling = scheduling
      return parts
    else:
      return content_api.ProcessorPart.from_function_response(
          name=call.name,
          response=parts,
          function_call_id=call.id,
          substream_name=substream_name,
          scheduling=scheduling,
          will_continue=will_continue,
          role='user',
      )

  async def _put_function_response_to_output_queue(
      self, call: genai_types.FunctionCall, fn: Callable[..., Any]
  ) -> None:
    """Puts the function call results to the output queue.

    Wraps all function call parts in a function response and puts them to the
    output queue. An EoT is also put to the output queue to request a model call
    in function calling loop. This function should only be called for async
    functions.

    Args:
      call: The function call to execute.
      fn: The function to execute.
    """
    async for fc_part in self._execute_function(call, fn):
      await self._output_queue.put(fc_part)

  async def _execute_function(
      self,
      call: genai_types.FunctionCall,
      fn: Callable[..., Any],
  ) -> AsyncIterable[content_api.ProcessorPart]:
    """Runs the function into an async Task."""
    converted_args = (
        _extra_utils.convert_number_values_for_dict_function_call_args(
            call.args
        )
    )
    try:
      if inspect.isasyncgenfunction(fn):
        # When the generator exits we automatically add a will_continue=False
        # Part. We send one to the downstream processor and also a separate
        # one for each of the reserved substreams as they have distinct
        # consumers.
        substreams_to_close = {self._substream_name}
        async for part_type in fn(**converted_args):
          response = self._to_function_response(part_type, call)
          substream = (
              response.substream_name
              if context_lib.is_reserved_substream(response.substream_name)
              else self._substream_name
          )
          substreams_to_close.add(substream)

          if response.function_response.will_continue is False:  # pylint: disable=g-bool-id-comparison
            substreams_to_close.discard(substream)
          else:
            response.function_response.will_continue = True

          yield response

        # Send a function response to indicate that the function call is
        # done (will_continue=False).
        for substream in substreams_to_close:
          yield content_api.ProcessorPart.from_function_response(
              name=call.name,
              function_call_id=call.id,
              response='',
              substream_name=substream,
              role='user',
          )
      elif inspect.iscoroutinefunction(fn):
        yield self._to_function_response(
            await fn(**converted_args), call, will_continue=None
        )
      else:
        yield self._to_function_response(
            await asyncio.to_thread(fn, **converted_args),
            call,
            will_continue=None,
        )
    except Exception as e:  # pylint: disable=broad-except
      yield content_api.ProcessorPart.from_function_response(
          name=call.name,
          function_call_id=call.id,
          response=(
              f'Failed to invoke function {fn.__name__}({converted_args}): {e}'
          ),
          role='user',
          substream_name=self._substream_name,
          is_error=True,
      )

  async def __call__(self, part: content_api.ProcessorPart) -> None:
    """Executes a function call and returns the result."""
    # The match() method ensures that the part is a function call.
    self._fn_state.fn_call_count += 1
    if self._fn_state.fn_call_count > self._fn_state.fn_call_count_limit:
      return

    part.substream_name = FUNCTION_CALL_SUBSTREAM_NAME
    call = part.function_call

    await self._output_queue.put(part)

    try:
      fn = self._fns[call.name]
    except KeyError:
      function_names = ', '.join(repr(fn) for fn in sorted(self._fns.keys()))
      await self._output_queue.put(
          content_api.ProcessorPart.from_function_response(
              name=call.name,
              function_call_id=call.id,
              response=(
                  f'Function {call.name} not found. Available functions:'
                  f' {function_names}.'
              ),
              role='user',
              substream_name=self._substream_name,
              will_continue=False if self._is_bidi_model else None,
              is_error=True,
          )
      )
      return

    if (
        inspect.iscoroutinefunction(fn)
        or inspect.isasyncgenfunction(fn)
        or self._is_bidi_model
    ):
      if call.id is None:
        call.id = self._create_task_id(call.name)
      # Immediately returns a function response with a std result. The
      # scheduling is silent.
      await self._output_queue.put(
          self._to_function_response(
              'Running in background.',
              call,
              scheduling=genai_types.FunctionResponseScheduling.SILENT,
              will_continue=True,
          )
      )
      # Run function in a separate task and inject the results back into the
      # output queue.
      self._add_task(
          call,
          fn,
      )
    else:
      # This is a sync function invoked by a non-bidi model. So we just run it
      # synchronously.
      async for fc_part in self._execute_function(call, fn):
        await self._output_queue.put(fc_part)



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""MIME types commonly used by Processors at the `Part` level.

Processors uses MIME types to express part modalities[^1].

If this file doesn't contain an appropriate MIME type you can pick one from
https://www.iana.org/assignments/media-types/media-types.xhtml
or define a custom */x-*.
"""

from typing import Any
from google.protobuf import message as pb_message


IMAGE_PNG = 'image/png'
IMAGE_JPEG = 'image/jpeg'
IMAGE_WEBP = 'image/webp'
IMAGE_HEIC = 'image/heic'
IMAGE_HEIF = 'image/heif'


INPUT_IMAGE_TYPES = [
    IMAGE_HEIF,
    IMAGE_HEIC,
    IMAGE_JPEG,
    IMAGE_PNG,
    IMAGE_WEBP,
]

AUDIO_AAC = 'audio/aac'
AUDIO_FLAC = 'audio/flac'
AUDIO_MP3 = 'audio/mp3'
AUDIO_M4A = 'audio/m4a'
AUDIO_MPEG = 'audio/mpeg'
AUDIO_MPGA = 'audio/mpga'
AUDIO_MP4 = 'audio/mp4'
AUDIO_OPUS = 'audio/opus'
AUDIO_PCM = 'audio/pcm'
AUDIO_WAV = 'audio/wav'
AUDIO_WEBM = 'audio/webm'

INPUT_AUDIO_TYPES = [
    AUDIO_AAC,
    AUDIO_FLAC,
    AUDIO_MP3,
    AUDIO_M4A,
    AUDIO_MPEG,
    AUDIO_MPGA,
    AUDIO_MP4,
    AUDIO_OPUS,
    AUDIO_PCM,
    AUDIO_WAV,
    AUDIO_WEBM,
]


VIDEO_MOV = 'video/mov'
VIDEO_MPEG = 'video/mpeg'
VIDEO_MPEGPS = 'video/mpegps'
VIDEO_MPG = 'video/mpg'
VIDEO_MP4 = 'video/mp4'
VIDEO_WEBM = 'video/webm'
VIDEO_WMV = 'video/wmv'
VIDEO_X_FLV = 'video/x-flv'
VIDEO_3GPP = 'video/3gpp'
VIDEO_QUICKTIME = 'video/quicktime'

INPUT_VIDEO_TYPES = [
    VIDEO_MOV,
    VIDEO_MPEG,
    VIDEO_MPEGPS,
    VIDEO_MPG,
    VIDEO_MP4,
    VIDEO_WEBM,
    VIDEO_WMV,
    VIDEO_X_FLV,
    VIDEO_3GPP,
    VIDEO_QUICKTIME,
]

TEXT_PDF = 'application/pdf'
TEXT_PLAIN = 'text/plain'
TEXT_CSV = 'text/csv'
TEXT_HTML = 'text/html'
TEXT_XML = 'text/xml'
TEXT_PYTHON = 'text/x-python'
TEXT_SCRIPT_PYTHON = 'text/x-script.python'
TEXT_JSON = 'application/json'
TEXT_ENUM = 'text/x.enum'

INPUT_TEXT_TYPES = [
    TEXT_PLAIN,
    TEXT_CSV,
    TEXT_HTML,
    TEXT_XML,
    TEXT_PYTHON,
    TEXT_SCRIPT_PYTHON,
    TEXT_JSON,
]

ALL_SUPPORTED_INPUT_TYPES = (
    INPUT_IMAGE_TYPES + INPUT_AUDIO_TYPES + INPUT_VIDEO_TYPES + INPUT_TEXT_TYPES
)

_PROTOBUF_MIMETYPE = 'application/x-protobuf'
_PROTOBUF_MIMETYPE_PREFIX = f'{_PROTOBUF_MIMETYPE}; type='
TEXT_EXCEPTION = 'text/x-exception'
CONTROL_TOKEN = 'application/x-control-token'
DIRECTIVE = 'application/x-directive'
FUNCTION_DECLARATION = 'application/x-function-declaration'


def is_text(mime: str) -> bool:
  """Returns whether the content is a human-readable text."""
  return (
      mime.lower() in INPUT_TEXT_TYPES
      or mime.lower().startswith('text/')
      or mime.lower().startswith(TEXT_JSON)
  )


def is_json(mime: str) -> bool:
  """Returns whether the content is a human-readable JSON."""
  return mime.lower().startswith(TEXT_JSON)


def is_dataclass(mime: str, json_dataclass: type[Any] | None = None) -> bool:
  """Returns whether the content is a dataclass."""
  type_name = json_dataclass.__name__ if json_dataclass else ''
  semi_colon_index = mime.lower().find('; type=')
  if semi_colon_index != -1:
    lower_index = semi_colon_index + len('; type=')
    mime_lower = mime[:lower_index].lower() + mime[lower_index:]
    return mime_lower.startswith(f'application/json; type={type_name}')
  else:
    return False


def is_image(mime: str) -> bool:
  """Returns whether the content is an image."""
  return (mime.lower() in INPUT_IMAGE_TYPES) or mime.lower().startswith(
      'image/'
  )


def is_video(mime: str) -> bool:
  """Returns whether the content is a video.

  Args:
    mime: The mime string.

  Returns:
    True if it is a video, False otherwise.
  """
  return (mime.lower() in INPUT_VIDEO_TYPES) or mime.lower().startswith(
      'video/'
  )


def is_audio(mime: str) -> bool:
  """Returns whether the content is audio."""
  return (mime.lower() in INPUT_AUDIO_TYPES) or mime.lower().startswith(
      'audio/'
  )


def is_streaming_audio(mime: str) -> bool:
  """Returns whether the content is streaming audio."""
  return mime.lower().startswith('audio/l16')


def is_wav(mime: str) -> bool:
  """Returns whether the content is a wav file."""
  return mime.lower() == AUDIO_WAV


def is_html(mime: str) -> bool:
  """Returns whether the content is HTML."""
  return mime.lower().startswith(TEXT_HTML)


def is_source_code(mime: str) -> bool:
  """Returns whether the content is a source code in some language."""
  # This list is incomplete and will be extended on as-needed basis.
  return mime.lower() in (
      'text/x-python',
      'application/x-latex',
      'text/x-c',
  )


def is_pdf(mime: str) -> bool:
  """Returns whether the content is a PDF."""
  return mime.lower() == TEXT_PDF


def is_csv(mime: str) -> bool:
  """Returns whether the content is a CSV file."""
  return mime.lower() == TEXT_CSV


def is_python(mime: str) -> bool:
  """Returns whether the content is python code."""
  return mime.lower() in [TEXT_PYTHON, TEXT_SCRIPT_PYTHON]


def is_exception(mime: str) -> bool:
  """Returns whether the content is an exception."""
  return mime.lower() == TEXT_EXCEPTION


def is_proto_message(
    mime: str, proto_message: type[pb_message.Message] | None = None
) -> bool:
  """Returns whether the mimetype is a proto message."""
  # We do case insensitive comparison for the `application/x-protobuf` part of
  # the mime type.
  if not mime.lower().startswith(_PROTOBUF_MIMETYPE):
    return False
  elif proto_message is None:
    return True
  elif not mime.lower().startswith(_PROTOBUF_MIMETYPE_PREFIX):
    return False
  else:
    mime = (
        mime[: len(_PROTOBUF_MIMETYPE_PREFIX)].lower()
        + mime[len(_PROTOBUF_MIMETYPE_PREFIX) :]
    )
    return mime == proto_message_mime_type(proto_message)


def proto_message_mime_type(proto_message: type[pb_message.Message]) -> str:
  """Returns the mimetype of a proto message."""
  return f'{_PROTOBUF_MIMETYPE_PREFIX}{proto_message.DESCRIPTOR.full_name}'



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

r"""Live commentator agent based on GenAI Processors.

Agent commentating on the video and audio stream from the default device inputs.
 This Agent can be run from a CLI directly.

## Setup

To install the dependencies for this script, run:

```
pip install pyaudio genai-processors
```

Before running this script, ensure the `GOOGLE_API_KEY` environment
variable is set to the api-key you obtained from Google AI Studio.

Important: **Use headphones**. This script uses the system default audio
input and output, which often won't include echo cancellation. So to prevent
the model from interrupting itself it is important that you use headphones.

## Run

To run the script:

```shell
python3 ./commentator_cli.py
```

The script takes a video-mode flag `--mode`, this can be "camera", "screen". The
default is "camera". To share your screen run:

```shell
python3 ./commentator_cli.py --mode=screen
```
"""

import argparse
import asyncio
import os

from absl import logging
from genai_processors.core import audio_io
from genai_processors.core import text
from genai_processors.core import video
import commentator
import pyaudio

# You need to define the API key in the environment variables.
# export GOOGLE_API_KEY=...
API_KEY = os.environ['GOOGLE_API_KEY']


async def run_commentator(video_mode: str) -> None:
  r"""Runs a live commentator in a CLI environment.

  The commentator is run from a CLI environment. The audio and video input and
  output are connected to the local machine's default input and output devices.


  Args:
    video_mode: The video mode to use for the video. Can be CAMERA or SCREEN.
  """
  pya = pyaudio.PyAudio()
  video_mode_enum = video.VideoMode(video_mode)
  input_processor = video.VideoIn(
      video_mode=video_mode_enum
  ) + audio_io.PyAudioIn(pya, use_pcm_mimetype=True)

  commentator_processor = commentator.create_live_commentator(API_KEY)

  consume_output = audio_io.PyAudioOut(pya)

  live_commentary_agent = (
      input_processor
      + commentator_processor
      + consume_output
  )

  print('Use ctrl+D to quit.')
  async for _ in live_commentary_agent(text.terminal_input()):
    pass


if __name__ == '__main__':
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '--mode',
      type=str,
      default='camera',
      help='pixels to stream from',
      choices=['camera', 'screen'],
  )
  parser.add_argument(
      '--debug',
      type=bool,
      default=False,
      help='Enable debug logging.',
  )
  args = parser.parse_args()
  if args.debug:
    logging.set_verbosity(logging.DEBUG)
  asyncio.run(run_commentator(video_mode=args.mode))



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for fetching documents mentioned in the part stream.

Referencing URLs of documents (web pages, images, PDFs) in a prompt is a
convenient way to provide rich context for a model. While registering `http.get`
as a tool is a more flexible and robust approach, it requires an extra model
call and a round trip for each document loaded. This approach offers a more
hardwired but faster alternative.

We split the responsibility for fetching documents and deciding what needs
fetching. A special `genai_processors.core.text.FetchRequest` part must be used
to explicitly reference the document to be fetched. Then `UrlFetch` processor
would replace such FetchRequest Parts with the actual content.

While it is very convenient to just mention URL as text in the prompt, it
becomes easy to trigger the fetch unintentionally and can even be dangerous. So
it should be applied closer to the UI where user journeys are more well defined.
For example parsing URLs directly pasted into a chat interface is probably
fine but recursively following URLs from uncontrolled sources is dangerous. For
extra safety you may want to require the URL be on its own line.
`genai_processors.core.text.UrlExtractor` is a processor for the task.

This process can be refined further: e.g. one can use a fast model
(gemini-flash-lite or gemma-nano) to decide whether the URL should be fetched
before passing the prompt to a larger LLM. This way we can reduce latency by
making decisions fast and fetching multiple documents in parallel.

You can also consider using an alternative implementation from
https://github.com/mbeacom/genai-processors-url-fetch/ which has additional
security features and supports markdown.
"""
from collections.abc import AsyncIterable

from genai_processors import content_api
from genai_processors import mime_types
from genai_processors import processor
from genai_processors.core import text
import httpx


_HEADER_PART = 'Fetch result for URL: '


class UrlFetch(processor.PartProcessor):
  """A processor that fetches documents by URLs.

  This processor replaces genai_processors.core.text.FetchRequest with the
  referenced content. It ignores anything that is not a FetchRequest.

  It is recommended to chain genai_processors.core.text.HtmlCleaner after
  UrlFetch to simplify the HTML pages before sending them to a model.
  """

  def __init__(
      self,
      timeout_seconds: int = 10,
  ):
    """Initializes the UrlFetch processor.

    Args:
      timeout_seconds: The timeout in seconds for the HTTP request. Default set
        to 10 seconds.
    """

    self._client = httpx.AsyncClient(
        follow_redirects=True,
        timeout=timeout_seconds,
    )

  def match(self, part: content_api.ProcessorPart) -> bool:
    return mime_types.is_dataclass(part.mimetype, text.FetchRequest)

  @processor.yield_exceptions_as_parts
  async def call(
      self, part: content_api.ProcessorPart
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    fetch_request = part.get_dataclass(text.FetchRequest)
    url = fetch_request.url
    yield _HEADER_PART + url + '\n'
    # Set stream=True to enable streaming the response body
    async with self._client.stream('GET', url, timeout=10) as response:
      # Raise an exception for bad status codes (4xx or 5xx)
      response.raise_for_status()

      # Iterate over the response chunks
      html_content = []
      async for chunk in response.aiter_text():
        html_content.append(chunk)
      yield content_api.ProcessorPart(
          ''.join(html_content), mimetype=mime_types.TEXT_HTML
      )



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Command Line Interface to test the Text-to-Speech GenAI Processor.

## Setup

To install the dependencies for this script, run:

```
pip install --upgrade google-cloud-texttospeech genai-processors
```

Before running this script, ensure the `GOOGLE_PROJECT_ID` environment
variable is set to your project_id (from Google Cloud Console).

## Run

To run the script:

```shell
python3 ./text_to_speech_cli.py
```

Enter the text you want to synthesize and press enter.
"""

import asyncio
import os
import time

from genai_processors.core import audio_io
from genai_processors.core import text
from genai_processors.core import text_to_speech
import pyaudio


# You need to define the project id in the environment variables.
# export GOOGLE_PROJECT_ID=...
GOOGLE_PROJECT_ID = os.environ['GOOGLE_PROJECT_ID']


async def run_tts() -> None:
  """Runs speech-to-text in a CLI environment."""

  pya = pyaudio.PyAudio()
  tts_processor = text_to_speech.TextToSpeech(
      project_id=GOOGLE_PROJECT_ID
  ) + audio_io.PyAudioOut(pya)

  print(
      f'{time.perf_counter()} - TTS Processor ready. Enter q to quit.\nNOTE:'
      ' finish all your input with a punctuation to indicate an end of'
      ' sentence. E.g.: "Hello, world!" or "hi."\nIMPORTANT: after 5 seconds'
      ' without activity (after the first sentence), the TTS will stop.'
  )
  print('Use ctrl+D to quit.')
  async for _ in tts_processor(text.terminal_input('message > ')):
    pass


if __name__ == '__main__':
  asyncio.run(run_tts())



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Low-level `ProcessorContent` caching implementation for processors.

Please don't use Caches directly, but instead wrap your processors with
processor.CachedPartProcessor.
"""

from collections.abc import Callable
import json
from typing import Optional, overload

import cachetools
from genai_processors import cache_base
from genai_processors import content_api
from genai_processors import mime_types
from typing_extensions import override
import xxhash


CacheMiss = cache_base.CacheMiss
CacheMissT = cache_base.CacheMissT
ProcessorContentTypes = content_api.ProcessorContentTypes
ProcessorContent = content_api.ProcessorContent


def default_processor_content_hash(
    processor_content_query: ProcessorContentTypes,
) -> str | None:
  """Creates a deterministic hash key from a ProcessorContent-like query.

  Serializes its parts to JSON and then hashes the result. The hash is
  order-sensitive with respect to the parts.

  Args:
    processor_content_query: The ProcessorContent-like object to hash.

  Returns:
    A string hash key.
  """
  content_obj = ProcessorContent(processor_content_query)

  # Do not cache content with errors.
  for part in content_obj.all_parts:
    if mime_types.is_exception(part.mimetype):
      return None

  raw_part_dicts = [part.to_dict() for part in content_obj.all_parts]
  for part in raw_part_dicts:
    part['metadata'].pop('capture_time', None)
  canonical_representation_str = json.dumps(raw_part_dicts, sort_keys=True)
  hasher = xxhash.xxh128()
  hasher.update(canonical_representation_str.encode('utf-8'))
  return hasher.hexdigest()


def prefixed_hash_fn(
    original_hash_fn: Callable[[content_api.ProcessorContentTypes], str | None],
    prefix: str,
) -> Callable[[content_api.ProcessorContentTypes], str | None]:
  """Returns a wrapped hash function that prepends a prefix to the key."""

  def _wrapper(query: content_api.ProcessorContentTypes) -> str | None:
    key = original_hash_fn(query)
    return f'{prefix}{key}' if key is not None else None

  return _wrapper


class InMemoryCache(cache_base.CacheBase):
  """An in-memory cache with TTL and size limits, specifically for caching `ProcessorContent`."""

  @overload
  def __init__(
      self,
      *,
      ttl_hours: float = 12,
      max_items: int = 1000,
      hash_fn: Callable[[ProcessorContentTypes], str | None] | None = None,
  ):
    ...

  @overload
  def __init__(
      self,
      *,
      base: 'InMemoryCache',
      hash_fn: Callable[[ProcessorContentTypes], str | None] | None = None,
  ):
    ...

  def __init__(
      self,
      *,
      ttl_hours: float = 12,
      max_items: int = 1000,
      base: Optional['InMemoryCache'] = None,
      hash_fn: Callable[[ProcessorContentTypes], str | None] | None = None,
  ):
    """Initializes the InMemoryCache for ProcessorContent.

    Args:
      ttl_hours: Time-to-live for cache items in hours.
      max_items: Maximum number of items in the cache. Must be positive.
      base: An instance of InMemoryCache with which this instance should share
        the cached items. Used to create a view of the cache `with_key_prefix`.
      hash_fn: Function to convert a ProcessorContentTypes query into a string
        key. If None, `default_processor_content_hash` is used. If it returns
        None, the item is considered not cacheable.
    """
    if max_items <= 0:
      raise ValueError('max_items must be positive, got: %d' % max_items)

    self._hash_fn = (
        hash_fn if hash_fn is not None else default_processor_content_hash
    )

    if base:
      self._cache = base._cache
    else:
      ttl_seconds = ttl_hours * 3600
      self._cache = cachetools.TTLCache(
          maxsize=max_items,
          ttl=ttl_seconds if ttl_seconds > 0 else float('inf'),
      )

  @property
  @override
  def hash_fn(self) -> Callable[[ProcessorContentTypes], str | None]:
    """Returns the function used to convert a query into a lookup key."""
    return self._hash_fn

  @override
  def with_key_prefix(self, prefix: str) -> 'InMemoryCache':
    """Creates a new InMemoryCache instance with its hash function wrapped to prepend the given prefix to generated string keys.

    The new instance uses the same TTL/max_items configuration but operates
    on a *new* underlying `cachetools.TTLCache`.

    Args:
      prefix: String to prepend to generated string keys.

    Returns:
      A new InMemoryCache instance with the given prefix.
    """
    return InMemoryCache(
        base=self,
        hash_fn=prefixed_hash_fn(self._hash_fn, prefix),
    )

  @override
  async def lookup(
      self,
      query: ProcessorContentTypes | None = None,
      *,
      key: str | None = None,
  ) -> ProcessorContent | CacheMissT:
    query_key = key if key is not None else self._hash_fn(query)
    if query_key is None:
      return CacheMiss

    cached_value = self._cache.get(query_key, CacheMiss)

    if cached_value is CacheMiss:
      return CacheMiss

    if not isinstance(cached_value, ProcessorContent):
      await self._remove_by_string_key(query_key)
      return CacheMiss

    return cached_value

  @override
  async def put(
      self,
      *,
      query: ProcessorContentTypes | None = None,
      key: str | None = None,
      value: ProcessorContentTypes,
  ) -> None:
    if self._cache.maxsize == 0:
      return

    query_key = key if key is not None else self._hash_fn(query)
    if query_key is None:
      return

    self._cache[query_key] = content_api.ProcessorContent(value)

  async def _remove_by_string_key(self, string_key: str) -> None:
    """Internal helper to remove by the actual string key."""
    if string_key in self._cache:
      del self._cache[string_key]

  @override
  async def remove(self, query: ProcessorContentTypes) -> None:
    query_key = self._hash_fn(query)
    if query_key is None:
      return
    await self._remove_by_string_key(query_key)



# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Syntax sugar for working with Processor `Content` and `Part` wrappers."""

from collections.abc import Callable, Iterable, Iterator
import copy
import dataclasses
import functools
import io
import itertools
import json
from typing import Any, AsyncIterable, AsyncIterator, Generator, Generic, Optional, TypeAlias, TypeVar, Union, overload

from absl import logging
from genai_processors import mime_types
from google.genai import types as genai_types
from google.protobuf import message as pb_message
from google.protobuf import timestamp
import PIL.Image


# bytes is used as a property in ProcessorPart, but pytype gets confused when
# checking function signatures. Use a type alias to avoid name conflict.
DataBytes: TypeAlias = bytes


class ProcessorPart:
  """A wrapper around `Part` with additional metadata.

  Represents a single piece of content that can be processed by an agentic
  system.

  Includes metadata such as the producer of the content, the substream the part
  belongs to, the MIME type of the content, and arbitrary metadata.
  """

  # GenAI SDK part representation of the ProcessorPart.
  _part: genai_types.Part | None = None
  # Metadata about the part. Can store arbitrary key/value pairs.
  _metadata: dict[str, Any] = {}
  # MIME type of the part.
  _mimetype: str | None = None
  # Role of the part.
  _role: str = ''
  # Substream name of the part.
  _substream_name: str = ''

  def __init__(
      self,
      value: 'ProcessorPartTypes',
      *,
      role: str = '',
      substream_name: str = '',
      mimetype: str | None = None,
      metadata: dict[str, Any] | None = None,
  ) -> None:
    """Constructs a ProcessorPart.

    Args:
      value: The content to use to construct the ProcessorPart.
      role: Optional. The producer of the content. In Genai models, must be
        either 'user' or 'model', but the user can set their own semantics.
        Useful to set for multi-turn conversations, otherwise can be empty.
      substream_name: (Optional) ProcessorPart stream can be split into multiple
        independent streams. They may have specific semantics, e.g. a song and
        its lyrics, or can be just alternative responses. Prefer using a default
        substream with an empty name. If the `ProcessorPart` is created using
        another `ProcessorPart`, this ProcessorPart inherits the existing
        substream_name, unless it is overridden in this argument.
      mimetype: Mime type of the data.
      metadata: (Optional) Auxiliary information about the part. If the
        `ProcessorPart` is created using another `ProcessorPart` or a
        `content_pb2.Part`, this ProcessorPart inherits the existing metadata,
        unless it is overridden in this argument.
    """
    super().__init__()
    self._metadata = {}

    match value:
      case genai_types.Part():
        self._part = value
      case genai_types.File():
        self._part = genai_types.Part(
            file_data=genai_types.FileData(
                display_name=value.name,
                mime_type=value.mime_type,
                file_uri=value.uri or '',
            ),
        )
        self._metadata['is_file'] = True
        mimetype = value.mime_type
      case ProcessorPart():
        self._part = value.part
        role = role or value.role
        substream_name = substream_name or value.substream_name
        mimetype = mimetype or value.mimetype
        self._metadata = value.metadata
        self._metadata.update(metadata or {})
      case str():
        self._part = genai_types.Part(text=value)
      case bytes():
        if not mimetype:
          raise ValueError(
              'MIME type must be specified when constructing a ProcessorPart'
              ' from bytes.'
          )
        if is_text(mimetype):
          self._part = genai_types.Part(text=value.decode('utf-8'))
        else:
          self._part = genai_types.Part.from_bytes(
              data=value, mime_type=mimetype
          )
      case PIL.Image.Image():
        if mimetype:
          # If the mimetype is explicitly specified, ensure it is an image.
          if not mimetype.startswith('image/'):
            raise ValueError(f"Can't convert image of mimetype {mimetype}.")
          suffix = mimetype[len('image/') :]
          # Ensure it matches the Image format.
          if value.format:
            if suffix != value.format.lower():
              raise ValueError(
                  f'The image format {value.format} does not match the'
                  f' mimetype {suffix}.'
              )
        else:
          # If no mimetype is specified, get it from the Image object.
          # If no format is provided, default to webp.
          suffix = value.format.lower() if value.format else 'webp'
          mimetype = f'image/{suffix}'
        bytes_io = io.BytesIO()
        value.save(bytes_io, suffix.upper())
        self._part = genai_types.Part.from_bytes(
            data=bytes_io.getvalue(), mime_type=mimetype
        )
      case _:
        raise ValueError(f"Can't construct ProcessorPart from {type(value)}.")

    self._role = role
    self._substream_name = substream_name
    self._metadata.update(metadata or {})

    # Set the MIME type.
    if mimetype:
      self._mimetype = mimetype
    # Otherwise, if MIME type is specified using inline data, use that.
    elif self._part.inline_data and self._part.inline_data.mime_type:
      self._mimetype = self._part.inline_data.mime_type
    elif self._part.function_call:
      # OSS library can't depend on protobuf, so we hardcode literal here.
      self._mimetype = (
          'application/x-protobuf;'
          ' type=google.ai.generativelanguage.v1main.FunctionCall'
      )
    elif self._part.function_response:
      # OSS library can't depend on protobuf, so we hardcode literal here.
      self._mimetype = (
          'application/x-protobuf;'
          ' type=google.ai.generativelanguage.v1main.FunctionResponse'
      )
    # Otherwise, if text is not empty, assume 'text/plain' MIME type.
    elif self._part and self._part.text:
      self._mimetype = 'text/plain'
    else:
      self._mimetype = ''

  def __repr__(self) -> str:
    optional_args = ''
    if self.substream_name:
      optional_args += f', substream_name={self.substream_name!r}'
    if self.metadata:
      optional_args += f', metadata={self.metadata}'.rstrip('\n')
    if self.role:
      optional_args += f', role={self.role!r}'
    return (
        f'ProcessorPart({self.part.to_json_dict()!r},'
        f' mimetype={self.mimetype!r}{optional_args})'
    )

  def __eq__(self, other: Any) -> bool:
    if not isinstance(other, ProcessorPart):
      return False

    return (
        self._part == other._part
        and self._role.lower() == other._role.lower()
        and self._substream_name.lower() == other._substream_name.lower()
        # Compare two dicts excluding 'capture_time' key.
        # Memory-efficient version.
        and set(['capture_time']).issuperset(
            k for (k, _) in self._metadata.items() ^ other._metadata.items()
        )
    )

  @property
  def part(self) -> genai_types.Part:
    """Returns the underlying Genai Part."""
    return self._part

  @property
  def file(self) -> genai_types.File | None:
    """Returns the underlying Genai File."""
    if (
        self._metadata.get('is_file', False)
        and self._part
        and self._part.file_data
    ):
      return genai_types.File(
          name=self._part.file_data.display_name,
          mime_type=self._part.file_data.mime_type,
          uri=self._part.file_data.file_uri,
      )
    else:
      return None

  @property
  def role(self) -> str:
    """Optional. The producer of the content.

    Useful to set for multi-turn conversations, otherwise can be left blank or
    unset.

    Default value is an empty string. It is up to the user to set their own
    semantics for the role.
    """
    return self._role

  @role.setter
  def role(self, value: str) -> None:
    self._role = value

  @property
  def bytes(self) -> bytes | None:
    """Returns part contents as bytes.

    Returns:
      Text encoded into bytes or bytes from inline data if the underlying part
      is a Blob.
    """
    if self.part.text is not None:
      return self.part.text.encode()
    if isinstance(self.part.inline_data, genai_types.Blob):
      return self.part.inline_data.data
    return None

  @property
  def substream_name(self) -> str:
    """Returns the stream this part belongs to.

    Empty for the default stream.
    """
    return self._substream_name

  @substream_name.setter
  def substream_name(self, value: str) -> None:
    self._substream_name = value

  @property
  def mimetype(self) -> str:
    """Returns part MIME type.

    Note: Empty MIME in the underlying `Part` is assumed to be text.
    """
    return self._mimetype or 'text/plain'

  @property
  def text(self) -> str:
    """Returns part text as string.

    Returns:
      The text of the part, which might be either in part.text or in
      part.inline_data.data if inline data contains a text file.

    Raises:
      ValueError if part is not of a text MIME type.
    """
    if not mime_types.is_text(self.mimetype):
      raise ValueError(f'Part is not text but {self._debug_string()}')
    if self.part.text:
      return self.part.text
    if self.part.inline_data:
      return self.part.inline_data.data.decode('utf-8')
    return ''

  @text.setter
  def text(self, value: str) -> None:
    """Sets part to a text part."""
    self._part = genai_types.Part(text=value)

  @property
  def metadata(self) -> dict[str, Any]:
    """Returns metadata."""
    return self._metadata

  @metadata.setter
  def metadata(self, value: dict[str, Any]) -> None:
    """Sets metadata."""
    self._metadata = value

  def get_metadata(self, key: str, default=None) -> Any:
    """Returns metadata for a given key."""
    return self._metadata.get(key, default)

  @property
  def function_call(self) -> genai_types.FunctionCall | None:
    """Returns function call."""
    return self.part.function_call

  @property
  def function_response(self) -> genai_types.FunctionResponse | None:
    """Returns function response."""
    return self.part.function_response

  @property
  def tool_cancellation(self) -> str | None:
    """Returns an ID of a function call to be cancelled.

    If the part is not a tool cancellation request, returns None.

    Returns:
      The id of the function call to be cancelled or None if this part is not a
      tool cancellation from the model.
    """
    if not self.part.function_response:
      return None
    if self.part.function_response.name != 'tool_cancellation':
      return None
    if not self.part.function_response.response:
      return None
    return self.part.function_response.response.get('function_call_id', None)

  T = TypeVar('T')

  def get_dataclass(self, json_dataclass: type[T]) -> T:
    """Returns representation of the Part as a given dataclass.

    Args:
      json_dataclass: A dataclass that can be converted to/from JSON.

    Returns:
      The dataclass representation of the Part.
    """
    if not mime_types.is_dataclass(self.mimetype):
      raise ValueError(f'Part is not a dataclass but {self._debug_string()}')
    try:
      # JSON conversions are provided by the dataclass_json decorator.
      return json_dataclass.from_json(self.text)  # pytype: disable=attribute-error
    except AttributeError as e:
      raise ValueError(
          f'{json_dataclass.__name__} is not a valid json dataclass'
      ) from e

  def get_proto_message(
      self, proto_message: type[pb_message.Message]
  ) -> pb_message.Message:
    """Returns representation of the Part as a given proto message."""
    if not mime_types.is_proto_message(self.mimetype, proto_message):
      raise ValueError(
          f'Part is not a {proto_message.DESCRIPTOR.name} proto message but'
          f' {self._debug_string()}.'
      )
    return proto_message.FromString(self.bytes)

  @property
  def pil_image(self) -> PIL.Image.Image:
    """Returns PIL.Image representation of the Part."""
    if not mime_types.is_image(self.mimetype):
      raise ValueError(f'Part is not an image but {self._debug_string()}.')
    bytes_io = io.BytesIO()
    if self.part.inline_data is not None:
      bytes_io.write(self.part.inline_data.data)
    bytes_io.seek(0)
    return PIL.Image.open(bytes_io)

  def _debug_string(self) -> str:
    if self.mimetype == mime_types.TEXT_EXCEPTION:
      return f'an exception: {self.text}'
    return self.mimetype

  # Class methods that make use of underlying Genai `Part` class methods.
  @classmethod
  def from_uri(
      cls, *, file_uri: str, mimetype: str, **kwargs
  ) -> 'ProcessorPart':
    """Constructs a ProcessorPart from URI & mimetype."""
    part = genai_types.Part.from_uri(file_uri=file_uri, mime_type=mimetype)
    return cls(part, **kwargs)

  @classmethod
  def from_function_call(
      cls, *, name: str, args: dict[str, Any], **kwargs
  ) -> 'ProcessorPart':
    """Constructs a ProcessorPart from function name & arguments."""
    part = genai_types.Part.from_function_call(name=name, args=args)
    return cls(part, **kwargs)

  @classmethod
  def from_function_response(
      cls,
      *,
      name: str | None = None,
      response: Union[Any, 'ProcessorContentTypes'],
      function_call_id: str | None = None,
      will_continue: bool | None = None,
      scheduling: genai_types.FunctionResponseScheduling | None = None,
      is_error: bool = False,
      **kwargs,
  ) -> 'ProcessorPart':
    """Constructs a ProcessorPart as a function response.

    Args:
      name: The name of the invoked function.
      response: The value returned by the function. It can be a
        JSON-serializable object or `ProcessorContentTypes`. If response is a
        JSON-serializable or is text-only it will be stored in the '.result'
        field of the function response. Otherwise it will be stored in `.parts`.
      function_call_id: The ID of the function call this is a response to if
        known.
      will_continue: Whether this is the last response from a generator
        function.
      scheduling: The scheduling policy for the function response. Controls
        whether the response generation will be triggered immediately or the
        function response is just added to the context.
      is_error: Whether the function response is an error. When it is an error,
        the response will be stored in the '.error' field in the function
        response.
      **kwargs: Additional keyword arguments to pass to the `ProcessorPart`
        constructor.

    Returns:
      A ProcessorPart representing the function response.
    """
    if is_error:
      response_key = 'error'
    else:
      response_key = 'result'
    function_response_args = dict(
        id=function_call_id,
        name=name,
        will_continue=will_continue,
        scheduling=scheduling,
    )
    # For maximal compatibility we first try to interpret response as JSON
    # serializable object. This is what tools in Gemini API return historically.
    try:
      function_response = genai_types.FunctionResponse(
          response={response_key: response}, **function_response_args
      )
      function_response.model_dump_json()
    except ValueError:
      # Response is not JSON serializable. Then try to construct content.
      response_content = ProcessorContent(response)
      try:
        function_response = genai_types.FunctionResponse(
            response={response_key: as_text(response_content, strict=True)},
            **function_response_args,
        )
      except ValueError:
        parts = [
            genai_types.FunctionResponsePart(
                inline_data=genai_types.FunctionResponseBlob(
                    data=part.bytes,
                    mime_type=part.mimetype,
                )
            )
            for part in response_content
        ]
        function_response = genai_types.FunctionResponse(
            parts=parts, **function_response_args
        )

    return cls(genai_types.Part(function_response=function_response), **kwargs)

  @classmethod
  def from_executable_code(
      cls, *, code: str, language: genai_types.Language, **kwargs
  ) -> 'ProcessorPart':
    """Constructs a ProcessorPart as an executable code part."""
    part = genai_types.Part.from_executable_code(code=code, language=language)
    return cls(part, **kwargs)

  @classmethod
  def from_code_execution_result(
      cls, *, outcome: genai_types.Outcome, output: str, **kwargs
  ) -> 'ProcessorPart':
    """Constructs a ProcessorPart as a code execution result part."""
    part = genai_types.Part.from_code_execution_result(
        outcome=outcome, output=output
    )
    return cls(part, **kwargs)

  @classmethod
  def from_bytes(
      cls, *, data: DataBytes, mimetype: str, **kwargs
  ) -> 'ProcessorPart':
    """Constructs a ProcessorPart as a blob."""
    part = genai_types.Part.from_bytes(data=data, mime_type=mimetype)
    return cls(part, **kwargs)

  @classmethod
  def from_proto_message(
      cls, *, proto_message: pb_message.Message, **kwargs
  ) -> 'ProcessorPart':
    """Constructs a ProcessorPart as a proto message."""
    kwargs['mimetype'] = mime_types.proto_message_mime_type(type(proto_message))
    return cls(
        proto_message.SerializeToString(),
        **kwargs,
    )

  @classmethod
  def from_tool_cancellation(
      cls, *, function_call_id: str, **kwargs
  ) -> 'ProcessorPart':
    """Constructs a ProcessorPart from a tool cancellation ID.

    The role is overridden to 'model'.

    Args:
      function_call_id: The id of the function call to be cancelled.
      **kwargs: Additional arguments for the ProcessorPart constructor.

    Returns:
      A ProcessorPart of type tool cancellation.
    """
    part = genai_types.Part.from_function_response(
        name='tool_cancellation',
        response={'function_call_id': function_call_id},
    )
    if 'role' in kwargs and kwargs['role'].lower() != 'model':
      logging.warning(
          'Role %s is not supported for tool cancellation.'
          ' Overriding it with the model role.',
          kwargs['role'],
      )
    extra_args = kwargs
    extra_args['role'] = 'model'
    return cls(part, **extra_args)

  @classmethod
  def from_dataclass(cls, dataclass: Any, **kwargs) -> 'ProcessorPart':
    """Constructs a ProcessorPart from a dataclass."""
    part = ProcessorPart(
        json.dumps(dataclasses.asdict(dataclass)),
        mimetype=f'application/json; type={type(dataclass).__name__}',
    )
    return cls(part, **kwargs)

  @classmethod
  def end_of_turn(
      cls, substream_name: str = '', metadata: dict[str, Any] | None = None
  ) -> 'ProcessorPart':
    metadata = metadata or {}
    metadata['turn_complete'] = True
    return ProcessorPart(
        '',
        role='user',
        substream_name=substream_name,
        metadata=metadata,
    )

  @classmethod
  def from_dict(cls, *, data: dict[str, Any]) -> 'ProcessorPart':
    """Deserializes a ProcessorPart from a JSON-compatible dictionary.

    This method reconstructs a ProcessorPart instance from a dictionary
    that was typically generated by the `to_dict()` method of another
    ProcessorPart instance.

    Args:
      data: A JSON-compatible dictionary containing the serialized data for the
        ProcessorPart.  It is expected to have the following keys:
        * 'part' (dict): A dictionary representing the underlying
          `google.genai.types.Part` object.
        * 'role' (str): The role of the part (e.g., 'user', 'model').
        * 'substream_name' (str): The substream name.
        * 'mimetype' (str): The MIME type of the part.
        * 'metadata' (dict[str, Any]): Auxiliary metadata.

    Returns:
      A new ProcessorPart instance.

    Raises:
      pydantic.ValidationError: If the `part` field in `data` is not a valid
        dictionary representation of a GenAI part.
      KeyError: If 'part' is missing from `data`.

    Example:
    ```py
    text_part = ProcessorPart("Hello", role="user")
    part_as_dict = text_part.to_dict()
    reconstructed = ProcessorPart.from_dict(data=part_as_dict)
    print(reconstructed)
    ```
    """  # fmt: skip
    return cls(
        genai_types.Part.model_validate(data['part']),
        role=data.get('role', ''),
        substream_name=data.get('substream_name', ''),
        mimetype=data.get('mimetype'),
        metadata=data.get('metadata'),
    )

  def to_dict(self, mode: str = 'json') -> dict[str, Any]:
    """Serializes this ProcessorPart to a JSON-compatible dictionary.

    The resulting dictionary can be used with `ProcessorPart.from_dict()`
    to reconstruct an equivalent ProcessorPart instance.

    Args:
      mode: The mode in which `to_dict` should run. If mode is 'json', the
        output will only contain JSON serializable types. Bytes are encoded as
        base64 strings. If mode is 'python', the output may contain
        non-JSON-serializable Python objects, i.e. bytes are left as is.

    Returns:
      A dictionary representing the ProcessorPart. It can be changed in place
      without affecting the original ProcessorPart instance.

      It is expected to have the following keys:
        * 'part' (dict): A dictionary representing the underlying
          `google.genai.types.Part` object.
        * 'role' (str): The role of the part (e.g., 'user', 'model').
        * 'substream_name' (str): The substream name.
        * 'mimetype' (str): The MIME type of the part.
        * 'metadata' (dict[str, Any]): Auxiliary metadata.


    Example:

    ```py
    text_part = ProcessorPart("Hello", role="user")
    part_as_dict = text_part.to_dict()
    print(part_as_dict)
    ```
    """
    return {
        'part': self.part.model_dump(mode=mode, exclude_none=True),
        'role': self.role,
        'substream_name': self.substream_name,
        'mimetype': self.mimetype,
        # Copy metadata to avoid caller modifying it in place.
        'metadata': copy.deepcopy(self.metadata),
    }

  def copy(self) -> 'ProcessorPart':
    """Returns a copy of the ProcessorPart.

    The internal part is not deep-copied, but the metadata is.

    Returns:
      A copy of the ProcessorPart.
    """
    return ProcessorPart(
        self.part,
        role=self.role,
        substream_name=self.substream_name,
        mimetype=self.mimetype,
        metadata=copy.deepcopy(self.metadata),
    )


T = TypeVar('T')


class _StreamOnce(Generic[T]):
  """An AsyncIterable that can be iterated over only once."""

  def __init__(self, stream: AsyncIterator[T]):
    self._stream = stream
    self._exhausted = False

  def __aiter__(self) -> AsyncIterator[T]:
    if self._exhausted:
      raise RuntimeError(
          'This ContentStream is backed by a generator and can only be'
          ' iterated over once.'
      )
    self._exhausted = True
    return self._stream


class ContentStream:
  """A syntax sugar mixin for streams of Content / Parts.

  AsyncIterable[Part], list[Part] or similar are common in LLM and agent
  interfaces. But there is mismatch between consumer and producer needs, which
  this class addresses.

  Models and agents need to work with multimodal streaming. They have to offer
  corresponding API to satisfy advanced users. But many consumers would benefit
  from simpler "I just need a string" interfaces. It works the other way too: if
  consumer can deal with streaming multimodal content, it should be able to
  ingest a plain string.

  ContentStream is an adaptor: it can be constructed from ProcessorContentTypes
  or Iterable/AsyncIterable/Collection of ProcessorContentTypes and then be
  consumed in a convient way through accessors like .text() which reduce the
  content to the given modality.

  Whether ContentStream can be iterated over only once (like generator) or
  multiple times (like a list) depends on the specific implementation.
  It is advised to allow reading content multiple times as this allows consumers
  to retry failures and tee response to multiple consumers. But if the
  ContentStream is backed by a generator, attempt to get content again (through
  any method) will raise a RuntimeError.

  Producers may chose to subclass ContentStream and provide additional methods.
  """

  def __init__(
      self,
      *,
      parts: Optional[AsyncIterable[ProcessorPart]] = None,
      parts_generator: Optional[AsyncIterator[ProcessorPart]] = None,
      content: Optional['ProcessorContentTypes'] = None,
  ):
    """Initializes the stream.

    Only one of parts, parts_generator or content can be set.
    Use `parts_generator` if content can be iterated over only once. `parts*` is
    more efficient than `content` if you already have ProcessorParts objects.

    Args:
      parts: Optimized version if you already have ProcessorParts. Constructs a
        stream that can be iterated over multiple times.
      parts_generator: Constructs a stream which can be iterated only once. The
        stream will raise a RuntimeError on consecutive access attempts. Use it
        if the underlying iterable discards the content as soon as it is
        consumed.
      content: Constructs the stream from a static content object.
    """
    if sum(x is not None for x in [content, parts, parts_generator]) > 1:
      raise ValueError(
          'At most one of content, content_generator, parts, parts_generator'
          ' can be provided.'
      )

    if content:
      parts = ProcessorContent(content)
    if parts_generator:
      parts = _StreamOnce(parts_generator)

    if parts:
      self._parts = parts

  def __aiter__(self) -> AsyncIterator[ProcessorPart]:
    """Returns an iterator that yields all genai.Parts from the stream.

    Consecutive calls to this method return independent iterators that start
    from the beginning of the stream. If the stream can only be iterated once,
    a RuntimeError will be risen on the second attempt.
    """
    return self._parts.__aiter__()

  async def text(
      self,
      *,
      # TODO(kibergus): Discuss whether default value should be true.
      strict: bool = False,
      substream_name: str | None = None,
  ) -> str:
    """Returns the stream contents as string.

    Args:
      strict: If True, unsupported content types will raise a ValueError.
        Otherwise, they will be ignored.
      substream_name: If set, only text parts with the given substream name will
        be returned.

    Returns:
      The text of the part.

    Raises:
      ValueError the underlying content contains non-text parts.
    """
    return as_text(
        await self.gather(), strict=strict, substream_name=substream_name
    )

  T = TypeVar('T')

  async def get_dataclass(self, json_dataclass: type[T]) -> T:
    """Returns representation of the stream as a given dataclass.

    Verifies that the stream contains just a single JSON Part (which is what
    models produce when constrained decoding is enabled) and instantiates the
    provided dataclass from it.

    Args:
      json_dataclass: A dataclass that can be converted to/from JSON.

    Returns:
      The dataclass representation of the Part.
    """
    content = await self.gather()
    if len(content) != 1:
      raise ValueError(
          'get_dataclass requires the stream to contain a single Part.'
      )
    return content[0].get_dataclass(json_dataclass)

  async def gather(self) -> 'ProcessorContent':
    """Returns all the contents of the stream as a list."""
    content = ProcessorContent()
    async for part in self:
      content += part
    return content

  def __await__(self) -> Generator[Any, None, None]:
    """Awaits until the stream is finished.

    Useful if we are not interested in the content itself, but in the side
    effect of the code that produces it.

    Returns:
      An awaitable that completes when the stream is finished.
    """

    async def await_parts() -> None:
      async for _ in self:
        pass

    return await_parts().__await__()

  # More accessor methods will be added here on as-needed basis.


class ProcessorContent(ContentStream):
  """A wrapper around `Content` with additional metadata.

  Serves as a convenience adaptor between various native representations and
  underlying data structures. ProcessorContent can be created from a string,
  image, ..., or a sequence of these.

  Users can narrow it down to a more convenient format using content_api.as_text
  or content_api.as_markdown. Or they can iterate over parts using .items()
  method.
  """

  _all_parts: list[ProcessorPart]

  def __init__(
      self,
      *parts: 'ProcessorContentTypes',
  ) -> None:
    """Constructs a new Content object from the given inputs."""
    super().__init__()
    self.replace_parts(*parts)

    self.as_text = functools.partial(as_text, self)
    self.as_text_with_reasoning = functools.partial(
        as_text_with_reasoning, self
    )
    self.as_images = functools.partial(as_images, self)

  def __iadd__(self, other: 'ProcessorContentTypes') -> 'ProcessorContent':
    """Appends other to the content."""
    if isinstance(other, ProcessorContent):
      self += other.all_parts
    elif isinstance(other, genai_types.Content):
      if other.parts:
        if other.role:
          parts = [ProcessorPart(part, role=other.role) for part in other.parts]
        else:
          parts = other.parts
        self += parts
    elif isinstance(other, ProcessorPart):
      self._all_parts.append(other)
    elif isinstance(other, ProcessorPartTypes):
      part = ProcessorPart(other)
      self._all_parts.append(part)
    elif isinstance(other, Iterable):
      for part in other:
        self += part
    else:
      raise ValueError(f"Can't append {type(other)} to ProcessorContent.")
    return self

  def __add__(self, other: 'ProcessorContentTypes') -> 'ProcessorContent':
    """Returns concatenation of two contents."""
    result = ProcessorContent()
    result += self
    result += other
    return result

  def __eq__(self, other: Any) -> bool:
    if isinstance(other, (list, tuple)):
      # This is very convenient for unit tests.
      try:
        other = ProcessorContent(other)
      except ValueError:
        return False
    if not isinstance(other, ProcessorContent):
      return False
    if len(self) != len(other):
      return False
    for lhs, rhs in zip(self, other):
      if lhs != rhs:
        return False
    return True

  def items(self) -> Iterator[tuple[str, ProcessorPart]]:
    """Yields tuples of mime_type and part.

    It is allowed to modify parts inplace except changing their IDs. Though
    bear in mind that like with most of Python containers that would change the
    part in all ProcessorContent containers which hold it.

    Yields:
      Tuples of mime_type, part.
    """
    for p in self.all_parts:
      yield p.mimetype, p

  def __iter__(self) -> Iterator[ProcessorPart]:
    """Yields each of the parts from this ProcessorContent.

    It is allowed to modify parts inplace.

    Bear in mind that like with most of Python containers that would change the
    part in all ProcessorContent containers which hold it.
    """
    for _, part in self.items():
      yield part

  def __aiter__(self) -> AsyncIterator[ProcessorPart]:
    """Returns an iterator that yields all genai.Parts from the content."""

    async def to_async_iter() -> AsyncIterator[ProcessorPart]:
      for item in self:
        yield item

    return to_async_iter()

  @property
  def all_parts(self) -> Iterator[ProcessorPart]:
    """Yields all ProcessorParts from this ProcessorContent."""
    yield from self._all_parts

  def replace_parts(self, *parts: 'ProcessorContentTypes') -> None:
    """Replaces this ProcessorContent's parts."""
    self._all_parts: list[ProcessorPart] = []
    for part in parts:
      self += part

  def __repr__(self) -> str:
    parts = ', '.join(repr(part) for part in self.all_parts)
    return f'ProcessorContent({parts})'

  def __len__(self) -> int:
    """Returns the number of parts in this ProcessorContent."""
    return len(self._all_parts)

  @overload
  def __getitem__(self, index: int) -> ProcessorPart:
    ...

  @overload
  def __getitem__(self, index: slice) -> list[ProcessorPart]:
    ...

  def __getitem__(
      self, index: int | slice
  ) -> ProcessorPart | list[ProcessorPart]:
    """Returns the part at the given index."""
    return self._all_parts[index]


# Prefer using ProcessorPart.end_of_turn() instead: it is too easy to mutate
# this global object.
END_OF_TURN = ProcessorPart.end_of_turn()


def is_end_of_turn(part: ProcessorPart) -> bool:
  """Returns whether the part is an end of turn event."""
  if part.get_metadata('turn_complete'):
    return True
  return False


# Types that can be converted to a ProcessorPart.
ProcessorPartTypes = (
    genai_types.Part
    | ProcessorPart
    | str
    | bytes
    | PIL.Image.Image
    | genai_types.File
)

# Types that can be appended to ProcessorContent.
ProcessorContentTypes = (
    ProcessorContent
    | ProcessorPartTypes
    | Iterable[ProcessorContent]
    | Iterable[ProcessorPartTypes]
    | genai_types.Content
    | Iterable[genai_types.Content]
)

# Helper functions for building content.

# Helper functions for mime type dispatching.
is_text = mime_types.is_text
is_json = mime_types.is_json
is_image = mime_types.is_image
is_video = mime_types.is_video
is_audio = mime_types.is_audio
is_streaming_audio = mime_types.is_streaming_audio
is_wav = mime_types.is_wav
is_source_code = mime_types.is_source_code
is_pdf = mime_types.is_pdf
is_csv = mime_types.is_csv
is_python = mime_types.is_python
is_dataclass = mime_types.is_dataclass
is_proto_message = mime_types.is_proto_message


def mime_type(part: ProcessorPart) -> str:
  """Returns the mimetype of the part."""
  return part.mimetype


def get_substream_name(
    part: ProcessorPart,
) -> str:
  """Returns the substream name of the part."""
  return part.substream_name


def group_by_mimetype(content: ProcessorContent) -> dict[str, ProcessorContent]:
  """Groups content by mimetype.

  The order of parts within each mimetype grouping is preserved, maintaining the
  same order as they appeared in the original input `content`.

  Args:
    content: The content to group.

  Returns:
    A dictionary mapping mimetypes to ProcessorContent objects, with the same
    order as in the original input `content`.
  """
  grouped_content = {}
  for mimetype, part in content.items():
    if mimetype not in grouped_content:
      grouped_content[mimetype] = ProcessorContent()
    grouped_content[mimetype] += part
  return grouped_content


# Functions that reduce ProcessorContent to well known formats.
def as_text(
    content: ProcessorContentTypes,
    *,
    strict: bool = False,
    substream_name: str | None = None,
) -> str:
  """Returns a text representation of the content.

  The returned text is a concatenation of all text parts in the content.

  Args:
    content: The content to process. This can be of various types as defined by
      `ProcessorContentTypes`.
    strict: If True, unsupported content types will raise a ValueError.
      Otherwise, they will be ignored.
    substream_name: If set, only text parts with the given substream name will
      be returned.
  """
  text_parts = []
  for mime, part in ProcessorContent(content).items():
    if substream_name is not None and part.substream_name != substream_name:
      continue
    if is_text(mime):
      text_parts.append(part.text)
    elif strict:
      raise ValueError(f'Unsupported content type {mime}.')

  return ''.join(text_parts)


def as_text_with_reasoning(
    content: ProcessorContentTypes,
    *,
    strict: bool = False,
) -> tuple[str, str]:
  """Returns a tuple of the final and reasoning text representing content.

  The returned tuple contains two elements:
    - The first element (index 0) is a string representing the main text
    extracted
      from the input `content`.
    - The second element (index 1) is a string representing the reasoning or
      thoughts associated with the input `content`.

  Args:
    content: The content to process. This can be of various types as defined by
      `ProcessorContentTypes`.
    strict: If True, unsupported content types will raise a ValueError.
      Otherwise, they will be ignored.

  Returns:
    A tuple containing two strings: (text, reasoning).
  """
  text_parts = []
  thought_parts = []
  for mime, p in ProcessorContent(content).items():
    if is_text(mime):
      if p.part.thought:
        thought_parts.append(p.text)
      else:
        text_parts.append(p.text)
    elif strict:
      raise ValueError(f'Unsupported content type {mime}.')

  return ''.join(text_parts), ''.join(thought_parts)


def _as_format_helper(
    content: ProcessorContentTypes,
    mime_check: Callable[[str], bool],
    ignore_unsupported_types: bool,
) -> list[ProcessorPart]:
  """Helper function to extract parts from the content based on MIME type."""
  if isinstance(content, ProcessorPart):
    # Fast path for singular parts.
    content = [content]
  elif not isinstance(content, ProcessorContent):
    content = ProcessorContent(content)

  parts = []
  for p in content:
    if mime_check(p.mimetype):
      parts.append(p)
    elif not ignore_unsupported_types:
      raise ValueError(f'Unsupported MIME type: {p.mimetype}.')
  return parts


def as_images(
    content: ProcessorContentTypes, *, ignore_unsupported_types: bool = False
) -> list[ProcessorPart]:
  """Returns the image parts from the content.

  Args:
    content: Input content.
    ignore_unsupported_types: By default if content contains non-image parts a
      ValueError would be risen. This argument allows to ignore such parts.

  Returns:
    A list of image parts, with the same order as in the input content.
  """
  return _as_format_helper(
      content, mime_types.is_image, ignore_unsupported_types
  )


def as_videos(
    content: ProcessorContentTypes, *, ignore_unsupported_types: bool = False
) -> list[ProcessorPart]:
  """Returns the video parts from the content.

  Args:
    content: Input content.
    ignore_unsupported_types: By default if content contains non-video parts a
      ValueError would be raised. This argument allows ingoring such parts.

  Returns:
    A list of video parts.
  """
  return _as_format_helper(
      content, lambda mime: mime.startswith('video/'), ignore_unsupported_types
  )


def to_genai_part(
    part_content: ProcessorPartTypes,
    mimetype: str | None = None,
) -> genai_types.Part:
  """Converts object of type `ProcessorPartTypes` to a Genai Part.

  Args:
    part_content: The content to convert.
    mimetype: (Optional) The mimetype of the content. Must be specified if
      part_content is bytes.

  Returns:
    The Genai Part representation of the content.
  """
  if isinstance(part_content, str):
    return genai_types.Part(text=part_content)
  elif isinstance(part_content, bytes):
    if mimetype is None:
      raise ValueError(
          'Mimetype must be specified for bytes to_genai_part conversion.'
      )
    p = ProcessorPart(part_content, mimetype=mimetype)
    return p.part
  elif isinstance(part_content, PIL.Image.Image):
    p = ProcessorPart(part_content)
    return p.part
  elif isinstance(part_content, ProcessorPart):
    return part_content.part
  elif isinstance(part_content, genai_types.Part):
    return part_content
  else:
    raise ValueError(
        f'Unsupported type for to_genai_part: {type(part_content)}'
    )


def to_genai_contents(
    content: ProcessorContentTypes,
) -> genai_types.ContentListUnion:
  """Converts a list of ProcessorParts into a list of Genai Content objects.

  Consecutive parts with the same role are grouped together into a single
  `genai_types.Content` object.

  Args:
    content: Processor content, e.g. a list of `ProcessorPartTypes`.

  Returns:
    A list of `genai_types.Content` objects, where each object represents
    content from a single role.
  """
  processor_content = ProcessorContent(content)
  contents = []
  for (role, _), parts_by_role_n_file in itertools.groupby(
      processor_content, lambda p: (p.role, p.file is not None)
  ):
    part_list = list(parts_by_role_n_file)
    if part_list and part_list[0].file is not None:
      contents += [p.file for p in part_list]
    else:
      content_parts = [p.part for p in part_list]
      contents.append(genai_types.Content(parts=content_parts, role=role))

  return contents



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Processors for handling constrained decoding and structured outputs."""

import dataclasses
import enum
import json
from typing import Any, AsyncIterable, Callable, Type, get_args, get_origin

from genai_processors import content_api
from genai_processors import processor


def _item_parser(
    item_type: Type[Any],
) -> Callable[[Any], content_api.ProcessorPart]:
  """Returns a parser for a single a Enum or Dataclass of type `item_type`."""
  if not isinstance(item_type, type):
    item_type = item_type.__class__

  if hasattr(item_type, 'from_dict') and dataclasses.is_dataclass(item_type):

    def parse_dataclass(data: Any) -> content_api.ProcessorPart:
      instance = item_type.from_dict(data)
      return content_api.ProcessorPart.from_dataclass(instance)

    return parse_dataclass

  elif issubclass(item_type, enum.Enum):

    def parse_enum(data: Any) -> content_api.ProcessorPart:
      instance = item_type(data)
      return content_api.ProcessorPart(instance.value)

    return parse_enum
  else:
    raise TypeError(
        f'{item_type.__name__} must be a dataclass with the'
        ' `dataclasses_json` mixin or an Enum.'
    )


class StructuredOutputParser(processor.Processor):
  """Buffers a stream of text parts to parse a JSON object or list.

  This processor is designed to be placed after a model processor that
  streams a JSON response. It consumes the entire input stream, buffers
  the text content, and then attempts to parse it as JSON.

  If the target `schema` is a list (e.g., `list[MyData]`), this processor will
  yield each item of the list as a separate `ProcessorPart` containing an
  instance of `MyData`. This is useful for chaining with PartProcessors that can
  then operate on each item concurrently.
  """

  def __init__(self, schema: type[Any]):
    """Initializes the processor.

    Args:
      schema: The schema to parse the JSON into. The schema must be a singleton
        enum or dataclass, or be a list of such types.
    """
    origin = get_origin(schema)
    is_list = origin is list

    if is_list:
      item_type = get_args(schema)[0]
      item_parser = _item_parser(item_type)

      async def parse_list(
          buffer: str,
      ) -> AsyncIterable[content_api.ProcessorPartTypes]:
        parsed_data = json.loads(buffer)
        if not isinstance(parsed_data, list):
          raise TypeError(
              'Model output was not a list, but expected'
              f' list[{item_type.__name__}].'
          )
        for item_value in parsed_data:
          yield item_parser(item_value)

      self._parser = parse_list
    else:
      item_type = schema
      item_parser = _item_parser(item_type)

      async def parse_single(
          buffer: str,
      ) -> AsyncIterable[content_api.ProcessorPartTypes]:
        try:
          parsed_data = json.loads(buffer)
        except json.JSONDecodeError:
          type_to_check = (
              item_type if isinstance(item_type, type) else item_type.__class__
          )
          if issubclass(type_to_check, enum.Enum):
            parsed_data = buffer
          else:
            raise
        yield item_parser(parsed_data)

      self._parser = parse_single

  @processor.yield_exceptions_as_parts
  async def call(
      self, content: AsyncIterable[content_api.ProcessorPart]
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    buffer = ''
    async for part in content:
      if content_api.is_text(part.mimetype):
        buffer += part.text
      else:
        yield part

    # Empty string is not a valid JSON. But we want to treat it as empty output.
    buffer = buffer.strip()
    if not buffer:
      return

    async for parsed_part in self._parser(buffer):
      yield parsed_part
